#include "cv.h"
#include "highgui.h"
#include "opencv.h"

result_ellipse_t *get_ellipses(char *filename, int threshold)
{
  IplImage *source = cvLoadImage(filename, 0);
  IplImage *working = cvCloneImage(source);

  CvMemStorage *stor = cvCreateMemStorage(0);
  CvSeq *contours =
    cvCreateSeq(CV_SEQ_ELTYPE_POINT, sizeof(CvSeq), sizeof(CvPoint) , stor);

  if(!contours)
    return 0;

  cvThreshold( source, working, threshold, 255, CV_THRESH_BINARY );

  cvFindContours( working, stor, &contours, sizeof(CvContour),
		  CV_RETR_LIST, CV_CHAIN_APPROX_NONE, cvPoint(0,0));

  int nr_of_contours = 0;
  {
    CvSeq *tmp;
    for(tmp = contours; tmp; nr_of_contours++, tmp = tmp->h_next);
  }

  result_ellipse_t *result = malloc(sizeof(result_ellipse_t));
  result->centers_x = malloc(nr_of_contours * sizeof(int));
  result->centers_y = malloc(nr_of_contours * sizeof(int));
  result->areas = malloc(nr_of_contours * sizeof(int));
  result->size = 0;

  for( ; contours ; contours = contours->h_next)
    {
      const int contours_size = contours->total;
      if( contours_size < 6)
	continue;

      // Alloc memory for contour point set.
      CvPoint *PointArray =
	(CvPoint *) malloc( contours_size * sizeof(CvPoint) );
      CvPoint2D32f *PointArray2D32f =
	(CvPoint2D32f *) malloc( contours_size * sizeof(CvPoint2D32f) );
      CvBox2D32f *box = (CvBox2D32f *) malloc(sizeof(CvBox2D32f));

      // Get contour point set.
      cvCvtSeqToArray(contours, PointArray, CV_WHOLE_SEQ);

      // Convert CvPoint set to CvBox2D32f set.
      int i;
      for(i = 0; i < contours->total; i++)
	{
	  PointArray2D32f[i].x = (float)PointArray[i].x;
	  PointArray2D32f[i].y = (float)PointArray[i].y;
	}

      cvFitEllipse(PointArray2D32f, contours_size, box);

      result->centers_x[result->size] = cvRound(box->center.x);
      result->centers_y[result->size] = cvRound(box->center.y);
      result->areas[result->size] = box->size.width*box->size.height;
      ++result->size;

      free(PointArray);
      free(PointArray2D32f);
      free(box);
    }

  return result;
}

int get_ellipses_x(result_ellipse_t *r, int n)
{
  return r->centers_x[n];
}

int get_ellipses_y(result_ellipse_t *r, int n)
{
  return r->centers_y[n];
}

int get_ellipses_area(result_ellipse_t *r, int n)
{
  return r->areas[n];
}

int get_number_of_ellipses(result_ellipse_t *r)
{
  return r->size;
}

void free_ellipses(result_ellipse_t *r)
{
  free(r->centers_x);
  free(r->centers_y);
  free(r);
}
