(MODULE
  TOOL1
  (WITH QOBISCHEME XLIB TOOL1LIB-HACK-TRACK TOOL1LIB-MOST MATLAB)
  (MAIN MAIN))

;;; LaHaShem HaAretz U'Mloah

;;; The origin is at the shoulder position. Bird's eye view.
;;; Z<---O
;;;      |
;;;      |
;;;      v
;;;      X
;;; Y points up. The arm points toward X.
;;; Theta is pitch. 0 is horizontal. Positive is upward.
;;; Phi is yaw. 0 is pointing along -Z. Positive is rightward.
;;; Psi is roll. It is fixed at 0 which is horizontal.

;;; Font height is hardwired.

;;; To Do
;;;  1. recording images and video
;;;  2. make everything cm or mm

(include "QobiScheme.sch")
(include "tool1.sch")
(include "c-externals.sc")
(include "c-macros.sch")

(set! *program* "tool1")
(set! *panic?* #f)			;debugging

;;; Macros

;;; Structures

(define-structure arm-configuration
 base-angle shoulder-angle elbow-angle hand-angle)

(define-structure square size x y contents name point)

;;; Variables

(define *camera* 0)
(define *number-of-cameras* 2)
(define *strings* #f)
(define *t* #f)
(define *end* #f)
(define *lines* #f)
(define *rules* #f)
(define *pathname* "learned-rules")
(define *square* #f)
(define *squares* #f)
(define *holding* #f)
(define *who-am-i* #f)
(define *bluetooth-address* #f)
(define *ellipses-to-display* #f)
(define *board-squares-to-display* #f)
(define *board-os-to-display* #f)
(define *lines-to-display* #f)
(define *game* #f)
(define *game-variant* #f)
(define *optimal?* #f)
(define *reconstruct?* #t)
(define *record?* #f)
(define *shoulder-relaxed?* #f)
(define *elbow-relaxed?* #f)
(define *head-relaxed?* #f)
(define *base* 1540)
(define *shoulder* 1900)
(define *elbow* 1100)
(define *hand* 1300)
(define *left-finger* 1800)
(define *right-finger* 1300)
(define *head* 1340)

;;; Parameters

(define *window-xs* '#(10 340 10 340))
(define *window-ys* '#(10 10 300 300))
(define *move-time* 750)		;ms
;;; Coefficients are us (servo pulse width) per radian.
(define *base-angle-coefficient* #f)
(define *shoulder-angle-coefficient* #f)
(define *elbow-angle-coefficient* #f)
(define *hand-angle-coefficient* #f)
;;; Offsets are in us (servo pulse width).
(define *base-angle-offset* #f)
(define *shoulder-angle-offset* #f)
(define *elbow-angle-offset* #f)
(define *hand-angle-offset* #f)
;;; These are in us (servo pulse width).
(define *open-left-finger* #f)
(define *open-right-finger* #f)
;;; These are in mm
(define *arm-ceiling* #f)
(define *board-height* #f)
(define *wrist-to-fingertip-pick-up* #f)
(define *wrist-to-fingertip-put-down* #f)
;;; Standard robot positions in us (servo pulse width).
(define *park-position* #f)
(define *reset-position* #f)
(define *center-head-position* #f)
;;; Finger positions in us (servo pulse width).
(define *closed-left-finger-X* #f)
(define *closed-right-finger-X* #f)
(define *closed-left-finger-O* #f)
(define *closed-right-finger-O* #f)
;;; Hard-coded positions (mm)
(define *square-xs* #f)
(define *square-ys* #f)
(define *staging-area-left* #f)
(define *staging-area-right* #f)
;;; For Pick Up and Put Down Motor Programs
(define *wait-x-coordinate* 70.6)	;mm
(define *cushion-ratio* 0.90)

;;; Procedures

;;; General Stuff

(define-c-external (c-cd pointer) int "cd")

(define (cd pathname)
 (unless (string? pathname) (panic "Argument to CD must be a string"))
 (let ((result (c-cd pathname)))
  (unless (zero? result) (panic "Cannot CD to directory"))
  #t))

(define (initialize-bluetooth-address!)
 (case *who-am-i*
  ((sahelanthropus) (set! *bluetooth-address* "00:06:66:00:9E:DF"))
  ((tchadensis) (set! *bluetooth-address* "00:06:66:00:16:B9"))
  ((orrorin) (set! *bluetooth-address* "00:06:66:00:9D:FC"))
  ((no-robot) (set! *bluetooth-address* #f))
  (else (fuck-up))))

(define (buffer-size)
 (case *number-of-cameras*
  ((1) 5200)
  ((2) 2600)
  ((4) 1300)
  (else (panic "Number of cameras must be 1, 2, or 4"))))

(define *unique-temporary-file-nr* 0)
(define (unique-temporary-file file)
  (set! *unique-temporary-file-nr* (+ *unique-temporary-file-nr* 1))
  (format #f
	  "~a-~a-~a-~a.~a"
	  (strip-extension file)
	  (getenv "USERNAME")
	  (getpid)
	  *unique-temporary-file-nr*
	  (extension file)))

;;; Matlab interface

(define *matlab-engine* #f)

(define (matlab-command string)
 (with-matlab-default-output-buffer
  *matlab-engine*
  (lambda (matlab-result-string)
   (matlab-eval-string *matlab-engine* (format #f "mresult=~a" string))
   (with-matlab-variable
    *matlab-engine*
    "mresult"
    (lambda (variable)
     (display (matlab-result-string)) (newline)
     (if (zero? variable)
	 (format #t "Command ~a failed~%" string)
	 (matlab->scheme variable)))))))

;;; Prolog interface

(define *prolog-input* #f)
(define *prolog-output* #f)

(define (start-prolog!)
 (let ((prolog-input-pathname
	(format #f "/tmp/prolog-output-~a-~a" (getenv "USERNAME") (getpid)))
       (prolog-output-pathname
	(format #f "/tmp/prolog-input-~a-~a" (getenv "USERNAME") (getpid))))
  (rm prolog-input-pathname)
  (rm prolog-output-pathname)
  (mkfifo prolog-input-pathname)
  (mkfifo prolog-output-pathname)
  (system
   (format #f "bash -c \"~~/imitate/swi-prolog/pl <~a 1>>~a 2>>/dev/null&\""
	   prolog-output-pathname prolog-input-pathname))
  (set! *prolog-input* (open-output-file prolog-output-pathname))
  (set! *prolog-output* (open-input-file prolog-input-pathname))
  #f))

(define (prolog->scheme string)
 (read-from-string
  (list->string
   (map (lambda (char)
	 (case char
	  ((#\[) #\()
	  ((#\]) #\))
	  ((#\,) #\space)
	  (else char)))
	(string->list string)))))

(define (prolog-command string)
 (display string *prolog-input*)
 (newline *prolog-input*)
 (flush-buffer *prolog-input*)
 (prolog->scheme (read-line *prolog-output*)))

(define (stop-prolog!)
 (close-output-port *prolog-input*)
 (let ((prolog-input-pathname
	(format #f "/tmp/prolog-output-~a-~a" (getenv "USERNAME") (getpid)))
       (prolog-output-pathname
	(format #f "/tmp/prolog-input-~a-~a" (getenv "USERNAME") (getpid))))
  (rm prolog-input-pathname)
  (rm prolog-output-pathname))
 (set! *prolog-input* #f)
 (set! *prolog-output* #f)
 #f)

(define (extract-clauses name all-lines)
  (let loop ((lines all-lines) (clauses '()))
    (if (null? lines)
	clauses
	(if (prefix? name (car lines))
	    (let ((dot-position
		   (position-if (lambda (x) (substring? "." x)) lines)))
	      (loop
	       (if dot-position
		   (sublist lines (+ 1 dot-position) (length lines))
		   '())
	       (append
		clauses
		`(,(sublist lines 0 (+ 1 dot-position))))))
	    (loop (cdr lines) clauses)))))

;;; Bluetooth interface

(define-c-external
 (c-bluetooth-connect int pointer) pointer "bluetooth_connect")
(define-c-external
 (c-bluetooth-disconnect pointer) void "bluetooth_disconnect")

(define *bluetooth-connection* #f)
(define *rfcomm-input* #f)
(define *rfcomm-output* #f)

(define (bluetooth-connect! bluetooth-address)
 (set! *bluetooth-connection* (c-bluetooth-connect 0 bluetooth-address))
 (when (or (not *bluetooth-connection*) (zero? *bluetooth-connection*))
  (format #f "Bluetooth connection failed~%")
  (fuck-up)))

(define (bluetooth-disconnect! connection)
 (c-bluetooth-disconnect connection)
 (set! *bluetooth-connection* #f))

(define (open-rfcomm! bluetooth-address)
 (bluetooth-connect! bluetooth-address)
 (set! *rfcomm-input* (open-input-file "/dev/rfcomm0"))
 (set! *rfcomm-output* (open-output-file "/dev/rfcomm0")))

(define (close-rfcomm!)
 (when *bluetooth-connection* (bluetooth-disconnect! *bluetooth-connection*))
 (close-input-port *rfcomm-input*)
 (close-output-port *rfcomm-output*)
 (set! *rfcomm-input* #f)
 (set! *rfcomm-output* #f))

;;; Display

(define (fixed n) (if (integer? n) (inexact->exact n) n))

(define (recache-strings!)
 (set! *strings*
       (map-n-vector
	(lambda (camera)
	 (list (format #f "T=~s, P=~s"
		       (fixed (/ (v4l-get-tilt camera) 100.0))
		       (fixed (- (/ (v4l-get-pan camera) 100.0) 90.0)))
	       (format #f "B=~s, C=~s, C=~s, W=~s"
		       (v4l-get-brightness camera)
		       (v4l-get-contrast camera)
		       (v4l-get-colour camera)
		       (v4l-get-whiteness camera))))
	*number-of-cameras*)))

(define (show-frame t)
 (xsetforeground *display* *thin-gc* (xcolor-pixel (second *background*)))
 (for-each-n
  (lambda (camera)
   (hack-track-show-frame camera t)
   (for-each (lambda (point)
	      (xdrawarc *display* *display-pane* *red-gc*
			(+ (- (x point) 3) (vector-ref *window-xs* camera))
			(+ (- (y point) 3) (vector-ref *window-ys* camera))
			6 6 0 (* 360 64)))
	     (vector-ref *ellipses-to-display* camera))
   (for-each (lambda (point)
	      (xdrawarc *display* *display-pane* *blue-gc*
			(+ (- (x point) 3) (vector-ref *window-xs* camera))
			(+ (- (y point) 3) (vector-ref *window-ys* camera))
			6 6 0 (* 360 64)))
	     (vector-ref *board-squares-to-display* camera))
   (for-each (lambda (point)
	      (xdrawarc *display* *display-pane* *orange-gc*
			(+ (- (x point) 3) (vector-ref *window-xs* camera))
			(+ (- (y point) 3) (vector-ref *window-ys* camera))
			6 6 0 (* 360 64)))
	     (vector-ref *board-os-to-display* camera))
   (for-each
    (lambda (line)
     (xdrawline *display* *display-pane* *green-gc*
		(+ (- (x (p line)) 3) (vector-ref *window-xs* camera))
		(+ (- (y (p line)) 3) (vector-ref *window-ys* camera))
		(+ (- (x (q line)) 3) (vector-ref *window-xs* camera))
		(+ (- (y (q line)) 3) (vector-ref *window-ys* camera))))
    (vector-ref *lines-to-display* camera)))
  *number-of-cameras*)
 (xsetforeground *display* *thin-gc* (xcolor-pixel (second *foreground*))))

(define (draw-square square)
 (define-region
  (square-x square) (square-y square)
  (square-size square) (square-size square)
  (lambda (x y)
   (message "")
   (set! *square* square)))
 (xcleararea *display* *display-pane*
	     (square-x square) (square-y square)
	     (square-size square) (square-size square)
	     0)
 (xdrawrectangle *display* *display-pane* *thin-gc*
		 (square-x square) (square-y square)
		 (square-size square) (square-size square))
 (case (square-contents square)
  ((x) (xdrawline *display* *display-pane* *thick-gc*
		  (+ (square-x square) 10)
		  (- (+ (square-y square) (square-size square)) 10)
		  (- (+ (square-x square) (square-size square)) 10)
		  (+ (square-y square) 10))
       (xdrawline *display* *display-pane* *thick-gc*
		  (+ (square-x square) 10)
		  (+ (square-y square) 10)
		  (- (+ (square-x square) (square-size square)) 10)
		  (- (+ (square-y square) (square-size square)) 10)))
  ((o) (xdrawarc *display* *display-pane* *thick-gc*
		 (+ (square-x square) 10) (+ (square-y square) 10)
		 (- (square-size square) 20) (- (square-size square) 20)
		 0 (* 360 64)))
  ((#f) #f)
  (else (fuck-up)))
 (xflush *display*))

(define (redraw)
 (when (eq? *mode* 'overlay) (show-frame *t*))
 (for-each-n
  (lambda (camera)
   (let ((strings (vector-ref *strings* camera)))
    (xdrawstring *display* *display-pane* *roman-gc*
		 (vector-ref *window-xs* camera)
		 (+ (vector-ref *window-ys* camera) 260)
		 (first strings) (string-length (first strings)))
    (xdrawstring *display* *display-pane* *roman-gc*
		 (vector-ref *window-xs* camera)
		 (+ (vector-ref *window-ys* camera) 275)
		 (second strings) (string-length (second strings)))))
  *number-of-cameras*)
 (unless (eq? *mode* 'overlay)
  (xdrawrectangle *display*
		  *display-pane*
		  *medium-gc*
		  (- (vector-ref *window-xs* *camera*) 5)
		  (- (vector-ref *window-ys* *camera*) 5)
		  330
		  250))
 (draw-strings-with-scroll-bar
  *lines*
  (+ (vector-ref *window-xs* 1) *width* 10)
  (- *display-pane-width* 10)
  5
  (+ (reduce max
	     (map-n (lambda (i) (vector-ref *window-ys* i))
		    *number-of-cameras*)
	     minus-infinity)
     260))
 (draw-strings-with-scroll-bar
  *rules*
  (+ (reduce max
	     (map (lambda (square) (+ (square-x square) (square-size square)))
		  *squares*)
	     minus-infinity)
     20)
  (- *display-pane-width* 10)
  (+ (reduce max
	     (map-n (lambda (i) (vector-ref *window-ys* i))
		    *number-of-cameras*)
	     minus-infinity)
     280)
  (- *display-pane-height* 5))
 (when *pathname*
  (xdrawstring
   *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 5)
   *pathname* (string-length *pathname*)))
 (for-each draw-square *squares*))

;;; Waiting

(define (wait-for-pan-tilt camera)
 ;; debugging
 (when #f
  (let loop ((i 0))
   (let ((status (v4l-pan-tilt-status camera)))
    (when (= i 20) (message "Time out waiting for pan/tilt") (abort))
    (when (zero? status)
     ;; Wait for pan/tilt.
     (interruptible-usleep 500000)
     (loop (+ i 1)))))))

(define (wait-for-next-frame frame)
 (let loop ((i 0))
  ;; hardwired
  ;; 200ms which is 5fps
  (when (= i 200) (message "Time out waiting for next frame") (abort))
  (when (some-n
	 (lambda (camera) (< (hack-track-get-current-frame camera) frame))
	 *number-of-cameras*)
   (interruptible-usleep 1000)
   (loop (+ i 1))))
 (unless (every-n
	  (lambda (camera) (= (hack-track-get-current-frame camera) frame))
	  *number-of-cameras*)
  (message
   (format #f "Skipped a frame ~s ~s"
	   (map-n (lambda (camera) (hack-track-get-current-frame camera))
		  *number-of-cameras*)
	   frame))
  (abort)))

;;; Robot Joint-Motion Primitives

(define (move-joint servo position)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#~s p~s t~s~a" servo position *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (cond ((= servo 1) (set! *shoulder-relaxed?* (zero? position)))
       ((= servo 2) (set! *elbow-relaxed?* (zero? position)))
       ((= servo 6) (set! *head-relaxed?* (zero? position)))))

(define (move-base base)
 (set! *base* base)
 (move-joint 0 *base*))

(define (move-shoulder shoulder)
 (set! *shoulder* shoulder)
 (move-joint 1 *shoulder*))

(define (move-elbow elbow)
 (set! *elbow* elbow)
 (move-joint 2 *elbow*))

(define (move-hand hand)
 (set! *hand* hand)
 (move-joint 3 *hand*))

(define (move-left-finger left-finger)
 (set! *left-finger* left-finger)
 (move-joint 4 *left-finger*))

(define (move-right-finger right-finger)
 (set! *right-finger* right-finger)
 (move-joint 5 *right-finger*))

(define (move-head head)
 (set! *head* head)
 (move-joint 6 *head*))

(define (relax-shoulder) (move-joint 1 0))

(define (unrelax-shoulder) (move-joint 1 *shoulder*))

(define (relax-elbow) (move-joint 2 0))

(define (unrelax-elbow) (move-joint 2 *elbow*))

(define (relax-shoulder-and-elbow)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#1 p0 #2 p0 t~s~a" *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (set! *shoulder-relaxed?* #t)
 (set! *elbow-relaxed?* #t))

(define (unrelax-shoulder-and-elbow)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#1 p~s #2 p~s t~s~a"
	 *shoulder* *elbow* *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (set! *shoulder-relaxed?* #f)
 (set! *elbow-relaxed?* #f))

(define (relax-head) (move-joint 6 0))

(define (unrelax-head) (move-joint 6 *head*))

(define (relax-shoulder-elbow-and-head)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#1 p0 #2 p0 #6 p0 t~s~a" *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (set! *shoulder-relaxed?* #t)
 (set! *elbow-relaxed?* #t)
 (set! *head-relaxed?* #t))

(define (unrelax-shoulder-elbow-and-head)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#1 p~s #2 p~s #6 p~s t~s~a"
	 *shoulder* *elbow* *head* *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (set! *shoulder-relaxed?* #f)
 (set! *elbow-relaxed?* #f)
 (set! *head-relaxed?* #f))

(define (move-arm base shoulder elbow)
 (set! *base* base)
 (set! *shoulder* shoulder)
 (set! *elbow* elbow)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#0 p~s #1 p~s #2 p~s t~s~a"
	 base shoulder elbow *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (set! *shoulder-relaxed?* (zero? shoulder))
 (set! *elbow-relaxed?* (zero? elbow)))

(define (move-arm-and-hand base shoulder elbow hand)
 (set! *base* base)
 (set! *shoulder* shoulder)
 (set! *elbow* elbow)
 (set! *hand* hand)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#0 p~s #1 p~s #2 p~s #3 p~s t~s~a"
	 base shoulder elbow hand *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (set! *shoulder-relaxed?* (zero? shoulder))
 (set! *elbow-relaxed?* (zero? elbow)))

(define (move-fingers left-finger right-finger)
 (set! *left-finger* left-finger)
 (set! *right-finger* right-finger)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#4 p~s #5 p~s t~s~a"
	 left-finger right-finger *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t))

(define (move-hand-and-fingers hand left-finger right-finger)
 (set! *hand* hand)
 (set! *left-finger* left-finger)
 (set! *right-finger* right-finger)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "#3 p~s #4 p~s #5 p~s t~s~a"
	 hand left-finger right-finger *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t))

(define (move-arm-hand-and-fingers
	 base shoulder elbow hand left-finger right-finger)
 (set! *base* base)
 (set! *shoulder* shoulder)
 (set! *elbow* elbow)
 (set! *hand* hand)
 (set! *left-finger* left-finger)
 (set! *right-finger* right-finger)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format
  *rfcomm-output* "#0 p~s #1 p~s #2 p~s #3 p~s #4 p~s #5 p~s t~s~a"
  base shoulder elbow hand left-finger right-finger *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (set! *shoulder-relaxed?* (zero? shoulder))
 (set! *elbow-relaxed?* (zero? elbow)))

(define (move-arm-hand-fingers-and-head
	 base shoulder elbow hand left-finger right-finger head)
 (set! *base* base)
 (set! *shoulder* shoulder)
 (set! *elbow* elbow)
 (set! *hand* hand)
 (set! *left-finger* left-finger)
 (set! *right-finger* right-finger)
 (set! *head* head)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format
  *rfcomm-output* "#0 p~s #1 p~s #2 p~s #3 p~s #4 p~s #5 p~s #6 p~s t~s~a"
  base shoulder elbow hand left-finger right-finger head *move-time* #\return)
 (flush-buffer *rfcomm-output*)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (set! *shoulder-relaxed?* (zero? shoulder))
 (set! *elbow-relaxed?* (zero? elbow))
 (set! *head-relaxed?* (zero? head)))

(define (read-nonblocking-string port)
 (let loop ((s '()))
  (if (char-ready? port)
      (loop (cons (read-char port) s))
      (list->string (reverse s)))))

(define (moving?)
 (hack-track-block)
 (hack-track-set-interrupts?! #f)
 (format *rfcomm-output* "q~aver~a" #\return #\return)
 (flush-buffer *rfcomm-output*)
 (let ((c (read-char *rfcomm-input*)))
  (read-line *rfcomm-input*)
  (hack-track-unblock)
  (hack-track-set-interrupts?! #t)
  (case c
   ((#\.) #f)
   ((#\+) #t)
   ((#\S) (read-nonblocking-string *rfcomm-input*) (moving?))
   (else (fuck-up)))))

(define (wait-until-servos-stop)
 (let loop ()
  (interruptible-usleep 200000)
  (let ((moving? (moving?))) (when moving? (loop)))))

(define (current-position)
 (list *base* *shoulder* *elbow* *hand* *left-finger* *right-finger*))

(define (park-arm) (apply move-arm-hand-and-fingers *park-position*))

(define (reset-arm) (apply move-arm-hand-and-fingers *reset-position*))

(define (center-head) (move-head *center-head-position*))

;;; Motor Control

(define (robot-parameter-pathname)
 (format #f "~a/imitate/tool1/~a.text"
	 (getenv "HOME")
	 (case *who-am-i*
	  ((sahelanthropus) "sahelanthropus")
	  ((tchadensis) "tchadensis")
	  ((orrorin) "orrorin")
	  (else (fuck-up)))))

(define (load-robot-parameters!)
 (let ((parameters (read-object-from-file (robot-parameter-pathname))))
  (set! *base-angle-coefficient* (list-ref parameters 0))
  (set! *shoulder-angle-coefficient* (list-ref parameters 1))
  (set! *elbow-angle-coefficient* (list-ref parameters 2))
  (set! *hand-angle-coefficient* (list-ref parameters 3))
  (set! *base-angle-offset* (list-ref parameters 4))
  (set! *shoulder-angle-offset* (list-ref parameters 5))
  (set! *elbow-angle-offset* (list-ref parameters 6))
  (set! *hand-angle-offset* (list-ref parameters 7))
  (set! *open-left-finger* (list-ref parameters 8))
  (set! *open-right-finger* (list-ref parameters 9))
  (set! *arm-ceiling* 274)
  (set! *board-height* 40)
  (set! *wrist-to-fingertip-pick-up* (list-ref parameters 10))
  (set! *wrist-to-fingertip-put-down* (list-ref parameters 11))
  (set! *park-position* (list-ref parameters 12))
  (set! *reset-position* (list-ref parameters 13))
  (set! *center-head-position* (list-ref parameters 14))
  (set! *closed-left-finger-X* (list-ref parameters 15))
  (set! *closed-right-finger-X* (list-ref parameters 16))
  (set! *closed-left-finger-O* (list-ref parameters 17))
  (set! *closed-right-finger-O* (list-ref parameters 18))
  (set! *square-xs* (list-ref parameters 19))
  (set! *square-ys* (list-ref parameters 20))
  (set! *staging-area-left* (list-ref parameters 21))
  (set! *staging-area-right* (list-ref parameters 22))))

(define (save-robot-parameters!)
 (write-object-to-file (list *base-angle-coefficient*
			     *shoulder-angle-coefficient*
			     *elbow-angle-coefficient*
			     *hand-angle-coefficient*
			     *base-angle-offset*
			     *shoulder-angle-offset*
			     *elbow-angle-offset*
			     *hand-angle-offset*
			     *open-left-finger*
			     *open-right-finger*
			     *wrist-to-fingertip-pick-up*
			     *wrist-to-fingertip-put-down*
			     *park-position*
			     *reset-position*
			     *center-head-position*
			     *closed-left-finger-X*
			     *closed-right-finger-X*
			     *closed-left-finger-O*
			     *closed-right-finger-O*
			     *square-xs*
			     *square-ys*
			     *staging-area-left*
			     *staging-area-right*)
		       (robot-parameter-pathname)))

(define (inverse-kinematics position)
 ;; position is a 3D coordinate in mm.
 ;; result is an arm configuration.
 ;; These offsets are in radians.
 ;; qobi: Need to document what all of the constants here are.
 (let* ((shoulder-angle-offset (asin (/ 20.1 152.2)))
	(elbow-angle-offset (asin (/ 88.9 152.2)))
	(xp (- (magnitude (vector (x position) (z position))) 19.5))
	(yp (- (y position) 79.6))
	(elbow-angle
	 (acos (- 1.0 (* 0.5 (sqr (/ (magnitude (vector xp yp)) 152.2))))))
	(shoulder-angle (- pi (+ (* 0.5 (- pi elbow-angle)) (atan yp xp)))))
  (make-arm-configuration (atan (z position) (x position))
			  (- shoulder-angle shoulder-angle-offset)
			  (- elbow-angle elbow-angle-offset)
			  (max 0.0 (- shoulder-angle elbow-angle)))))

(define (radians->pulse-width x m b) (inexact->exact (round (+ (* m x) b))))

(define (arm-configuration-radians->pulse-width arm-configuration)
 (make-arm-configuration
  (radians->pulse-width (arm-configuration-base-angle arm-configuration)
			*base-angle-coefficient*
			*base-angle-offset*)
  (radians->pulse-width (arm-configuration-shoulder-angle arm-configuration)
			*shoulder-angle-coefficient*
			*shoulder-angle-offset*)
  (radians->pulse-width (arm-configuration-elbow-angle arm-configuration)
			*elbow-angle-coefficient*
			*elbow-angle-offset*)
  (radians->pulse-width (arm-configuration-hand-angle arm-configuration)
			*hand-angle-coefficient*
			*hand-angle-offset*)))

(define (move-end-effector position piece open?)
 ;; This moves the end effector to a 3D position in mm.
 (let ((arm-configuration (arm-configuration-radians->pulse-width
			   (inverse-kinematics position))))
  (move-arm-hand-and-fingers
   (arm-configuration-base-angle arm-configuration)
   (arm-configuration-shoulder-angle arm-configuration)
   (arm-configuration-elbow-angle arm-configuration)
   (arm-configuration-hand-angle arm-configuration)
   (if open?
       *open-left-finger*
       (cond ((eq? piece 'o) *closed-left-finger-O*)
	     ((eq? piece 'x) *closed-left-finger-X*)
	     (else (fuck-up))))
   (if open?
       *open-right-finger*
       (cond ((eq? piece 'o) *closed-right-finger-O*)
	     ((eq? piece 'x) *closed-right-finger-X*)
	     (else (fuck-up)))))))

(define (move-segmented-end-effector position piece open?)
 ;; This moves the end effector to a 3D position in mm.
 (let ((arm-configuration (arm-configuration-radians->pulse-width
			   (inverse-kinematics position))))
  (move-hand-and-fingers (arm-configuration-hand-angle arm-configuration)
			 (if open?
			     *open-left-finger*
			     (cond ((eq? piece 'o) *closed-left-finger-O*)
				   ((eq? piece 'x) *closed-left-finger-X*)
				   (else (fuck-up))))
			 (if open?
			     *open-right-finger*
			     (cond ((eq? piece 'o) *closed-right-finger-O*)
				   ((eq? piece 'x) *closed-right-finger-X*)
				   (else (fuck-up)))))
  (move-arm (arm-configuration-base-angle arm-configuration)
	    (arm-configuration-shoulder-angle arm-configuration)
	    (arm-configuration-elbow-angle arm-configuration))))

(define (check-arm-state)
 (when (or *shoulder-relaxed?* *elbow-relaxed?*)
  (message "Arm is relaxed")
  (abort)))

(define (pick-up-motor-program position piece)
 ;; This picks up an object at position (x,y,z) on the ground. (mm)
 (check-arm-state)
 (let* ((x (x position))
	(y (y position))
	(z (z position))
	(hover (+ *wrist-to-fingertip-pick-up* y)))
  (move-segmented-end-effector
   (vector *wait-x-coordinate* *arm-ceiling* 0) piece #t)
  (wait-until-servos-stop)
  (move-segmented-end-effector
   (vector (* x *cushion-ratio*) *arm-ceiling* (* z *cushion-ratio*)) piece #t)
  (wait-until-servos-stop)
  (move-segmented-end-effector (vector x hover z) piece #t)
  (wait-until-servos-stop)
  (move-segmented-end-effector (vector x hover z) piece #f)
  (wait-until-servos-stop)
  (move-segmented-end-effector
   (vector (* x *cushion-ratio*) *arm-ceiling* (* z *cushion-ratio*)) piece #f)
  (wait-until-servos-stop)))

(define (put-down-motor-program position piece)
 ;; This picks up an object at position (x,y,z) on the ground. (mm)
 (check-arm-state)
 (let* ((x (x position))
	(y (y position))
	(z (z position))
	(hover (+ *wrist-to-fingertip-put-down* y)))
  (move-segmented-end-effector
   (vector (* x *cushion-ratio*) *arm-ceiling* (* z *cushion-ratio*)) piece #f)
  (wait-until-servos-stop)
  (move-segmented-end-effector (vector x hover z) piece #f)
  (wait-until-servos-stop)
  (move-segmented-end-effector (vector x hover z) piece #t)
  (wait-until-servos-stop)
  (move-segmented-end-effector
   (vector (* x *cushion-ratio*) *arm-ceiling* (* z *cushion-ratio*)) piece #t)
  (wait-until-servos-stop)
  (move-segmented-end-effector
   (vector *wait-x-coordinate* *arm-ceiling* 0) piece #t)
  (wait-until-servos-stop)))

(define (square->position square)
 (cond ((eq? (first (square-name square)) 'board)
	(vector (matrix-ref *square-xs*
			    (second (square-name square))
			    (third (square-name square)))
		*board-height*
		(matrix-ref *square-ys*
			    (second (square-name square))
			    (third (square-name square)))))
       ((eq? (first (square-name square)) 'staging)
	(vector-ref (case (second (square-name square))
		     ((left) *staging-area-left*)
		     ((right) *staging-area-right*)
		     (else (fuck-up)))
		    (third (square-name square))))
       (else (fuck-up))))

(define (pick-up square)
 (when *bluetooth-address* (check-arm-state))
 (when *holding* (message "You are already holding something") (abort))
 (set! *holding* (square-contents square))
 (when *bluetooth-address*
  (pick-up-motor-program (square->position square) (square-contents square)))
 (unless *reconstruct?*
  (set-square-contents! square #f)
  (redraw-display-pane)))

(define (put-down square)
 (when *bluetooth-address* (check-arm-state))
 (unless *holding* (message "You are not holding anything") (abort))
 (unless *reconstruct?* (set-square-contents! square *holding*))
 (when *bluetooth-address*
  (put-down-motor-program (square->position square) *holding*))
 (set! *holding* #f)
 (unless *reconstruct?* (redraw-display-pane)))

;;; Ellipse finder

(c-include "cv.h")
(c-include "opencv.h")

(define get-ellipses
 (c-function pointer ("get_ellipses" string int)))
(define free-ellipses
 (c-function void ("free_ellipses" pointer)))
(define get-ellipses-x (c-function int ("get_ellipses_x" pointer int)))
(define get-ellipses-y (c-function int ("get_ellipses_y" pointer int)))
(define get-ellipses-area (c-function int ("get_ellipses_area" pointer int)))
(define get-number-of-ellipses (c-function int ("get_number_of_ellipses" pointer)))

(define *board-square-distance* #f)
(define *board-square-ellipse-finder-threshold* #f)
(define *board-square-ellipse-min-area* #f)
(define *board-square-ellipse-max-area* #f)
(define *board-square-ellipse-min-x* #f)
(define *board-square-ellipse-min-y* #f)
(define *o-ellipse-finder-threshold* #f)
(define *o-ellipse-finder-min-area* #f)
(define *o-ellipse-finder-max-area* #f)

(define (load-ellipse-finder-parameters!)
 (let ((parameters
	(read-object-from-file
	 (format #f "~a/imitate/tool1/ellipse-finder-parameters.text"
		 (getenv "HOME")))))
  (set! *board-square-distance* (first parameters))
  (set! *board-square-ellipse-finder-threshold* (second parameters))
  (set! *board-square-ellipse-min-area* (third parameters))
  (set! *board-square-ellipse-max-area* (fourth parameters))
  (set! *board-square-ellipse-min-x* (fifth parameters))
  (set! *board-square-ellipse-min-y* (sixth parameters))
  (set! *o-ellipse-finder-threshold* (seventh parameters))
  (set! *o-ellipse-finder-min-area* (eighth parameters))
  (set! *o-ellipse-finder-max-area* (ninth parameters))))

(define (pnm->ellipses pnm threshold)
 (let ((pathname
	(format #f "/tmp/ellipses-~a-~a.pgm" (getenv "USERNAME") (getpid))))
  (write-pnm pnm pathname)
  (let* ((es (get-ellipses pathname threshold))
	 (ellipses (map-n (lambda (n)
			   (vector (get-ellipses-x es n)
				   (get-ellipses-y es n)
				   (get-ellipses-area es n)))
			  (get-number-of-ellipses es))))
   (free-ellipses es)
   (rm pathname)
   ellipses)))

(define (ellipse-size e) (vector-ref e 2))

(define (piece-in-range checking-position board-position)
 (let ((horizontal-range 4))
  (if (and (>= (y checking-position)
	       (- (y board-position) *board-square-distance*))
	   (<= (y checking-position) (y board-position)))
      (and (<= (x checking-position) (+ (x board-position) horizontal-range))
	   (>= (x checking-position) (- (x board-position) horizontal-range)))
      #f)))

(define (empty-square-in-range new-empty-square old-empty-square)
 (let ((range 5))
  (<= (distance (vector (x new-empty-square) (y new-empty-square))
		(vector (x old-empty-square) (y old-empty-square)))
      range)))

(define (pnm->ellipses-of-area pnm threshold min max)
 (remove-if (lambda (e) (or (< (ellipse-size e) min) (> (ellipse-size e) max)))
	    (pnm->ellipses pnm threshold)))

(define (pnm->board-square-ellipses pnm)
 (remove-if (lambda (ellipse)
	     (or (< (x ellipse) *board-square-ellipse-min-x*)
		 (< (y ellipse) *board-square-ellipse-min-y*)
		 (> (x ellipse) (- *width* *board-square-ellipse-min-x*))
		 (> (y ellipse) (- *height* *board-square-ellipse-min-y*))))
	    (pnm->ellipses-of-area pnm
				   *board-square-ellipse-finder-threshold*
				   *board-square-ellipse-min-area*
				   *board-square-ellipse-max-area*)))

(define (similar-ellipses? ellipse1 ellipse2)
 (let ((distance-variance 10)
       (area-variance 60))
  (and (<= (distance (vector (x ellipse1) (y ellipse1))
		     (vector (x ellipse2) (y ellipse2)))
	   distance-variance)
       (<= (abs (- (z ellipse1) (z ellipse2))) area-variance))))

(define (distinct-ellipses ellipses)
 (if (null? ellipses)
     '()
     (if (find-if (lambda (x) (similar-ellipses? (first ellipses) x))
		  (rest ellipses))
	 (distinct-ellipses (rest ellipses))
	 (cons (first ellipses) (distinct-ellipses (rest ellipses))))))

(define (attach-count ellipses)
 (if (null? ellipses) '()
     (cons (list (first ellipses) 1) (attach-count (rest ellipses)))))

(define (merge-ellipses consolidated-ellipses new-ellipses)
 (let* ((new-consolidated-ellipses
	(map (lambda (e1)
	      (let ((common (find-if
			     (lambda (e2)
			      (similar-ellipses? (first e2) (first e1)))
			     new-ellipses)))
	       (if common
		   (list (first common) (+ (second common) (second e1)))
		   e1)))
	     consolidated-ellipses))
       (non-common-new-ellipses
	(reduce intersection
		(map
		 (lambda (p)
		  (removep
		   (lambda (e1 e2) (similar-ellipses? (first e1) (first e2)))
		   p
		   new-ellipses))
		new-consolidated-ellipses)
		'())))
  (append new-consolidated-ellipses non-common-new-ellipses)))


(define (invariants pnm l-threshold u-threshold step consolidated-ellipses)
 (if (> l-threshold u-threshold)
     (map first
	  (remove #f
		  (map (lambda (x) (if (< (second x) 4) #f x))
		       consolidated-ellipses)))
     (let* ((new-ellipses (pnm->ellipses-of-area pnm
						 l-threshold
						 *o-ellipse-finder-min-area*
						 *o-ellipse-finder-max-area*))
	    (new-distinct-ellipses (distinct-ellipses new-ellipses))
	    (new-ellipses-with-count (attach-count new-distinct-ellipses))
	    (merged-ellipses (merge-ellipses consolidated-ellipses
					     new-ellipses-with-count)))
      (invariants pnm (+ l-threshold step) u-threshold step
		  (if (null? merged-ellipses)
		      new-ellipses-with-count
		      merged-ellipses)))))




;;; Squares

(define (create-square x y contents name)
 (make-square 40 x y contents name #f))

(define (board-square-at i j)
 (find-if (lambda (square) (equal? (square-name square) `(board ,i ,j)))
	  *squares*))

(define (left-staging-square-at i)
 (find-if (lambda (square) (equal? (square-name square) `(staging left ,i)))
	  *squares*))

(define (right-staging-square-at i)
 (find-if (lambda (square) (equal? (square-name square) `(staging right ,i)))
	  *squares*))

;;; Board reconstruction

(define (split-into n l)
 (if (< (length l) n)
     (list l)
     (cons (sublist l 0 n) (split-into n (sublist l n (length l))))))

(define (find-board-in-pnm! pnm camera)
 (set! *ellipses-to-display* (make-vector *number-of-cameras* '()))
 (set! *board-squares-to-display* (make-vector *number-of-cameras* '()))
 (set! *board-os-to-display* (make-vector *number-of-cameras* '()))
 (set! *lines-to-display* (make-vector *number-of-cameras* '()))
 (let ((points (map (lambda (e) (vector (x e) (y e)))
		    (pnm->board-square-ellipses pnm))))
  (vector-set! *ellipses-to-display* camera
	       (pnm->ellipses pnm *board-square-ellipse-finder-threshold*))
  (vector-set! *board-squares-to-display* camera
	       (pnm->board-square-ellipses pnm))
  (unless (= (length points) 9)
   (message "Cannot find exactly nine board squares")
   (abort))
  (for-each-indexed
   (lambda (points i)
    (for-each-indexed (lambda (point j)
		       (set-square-point! (board-square-at j i) point))
		      (sort points < y)))
   (split-into 3 (sort points < x)))))

(define (update-staging-area! board)
 (let ((number-of-xs-not-to-put-in-staging-area
	(count 1 (reduce append board '())))
       (number-of-os-not-to-put-in-staging-area
	(count -1 (reduce append board '()))))
  (for-each-n
   (lambda (i)
    (set-square-contents!
     (left-staging-square-at i)
     (if (< i number-of-xs-not-to-put-in-staging-area) #f 'x)))
   5)
  (for-each-n
   (lambda (i)
    (set-square-contents!
     (right-staging-square-at i)
     (if (< i number-of-os-not-to-put-in-staging-area) #f 'o)))
   5)))

(define (mean-hue-in-box ppm xmin ymin xmax ymax)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm))
       (hues '()))
  (do ((i ymin (+ i 1))) ((= i ymax))
   (do ((j xmin (+ j 1))) ((= j xmax))
    (set! hues (cons (vector-ref (rgb->hsv (vector (matrix-ref red i j)
						   (matrix-ref green i j)
						   (matrix-ref blue i j)))
				 0)
		     hues))))
  (mean-angle hues)))

(define (mean-saturation-in-box ppm xmin ymin xmax ymax)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm))
       (saturations '()))
  (do ((i ymin (+ i 1))) ((= i ymax))
   (do ((j xmin (+ j 1))) ((= j xmax))
    (set! saturations
	  (cons (vector-ref (rgb->hsv (vector (matrix-ref red i j)
					      (matrix-ref green i j)
					      (matrix-ref blue i j)))
			    1)
		saturations))))
  (/ (reduce + saturations 0) (length saturations))))

(define (mean-value-in-box ppm xmin ymin xmax ymax)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm))
       (values '()))
  (do ((i ymin (+ i 1))) ((= i ymax))
   (do ((j xmin (+ j 1))) ((= j xmax))
    (set! values
	  (cons (vector-ref (rgb->hsv (vector (matrix-ref red i j)
					      (matrix-ref green i j)
					      (matrix-ref blue i j)))
			    2)
		values))))
  (/ (reduce + values 0) (length values))))

(define (reconstruct-board! ppm camera)
 (unless (every square-point
		(remove-if (lambda (square)
			    (not (equal? (first (square-name square)) 'board)))
			   *squares*))
  (message "You must first click on Find Board")
  (abort))
 (set! *ellipses-to-display* (make-vector *number-of-cameras* '()))
 (set! *board-squares-to-display* (make-vector *number-of-cameras* '()))
 (set! *board-os-to-display* (make-vector *number-of-cameras* '()))
 (set! *lines-to-display* (make-vector *number-of-cameras* '()))
 (for-each
  (lambda (square)
   (let ((mean-value
	  (mean-value-in-box ppm
			     (- (x (square-point square)) 10)
			     (- (y (square-point square)) 10)
			     (+ (x (square-point square)) 10)
			     (+ (y (square-point square)) 10))))
    (set-square-contents!
     square
     (cond
      ((> mean-value 55) #f)
      ((< mean-value 25) 'o)
      (else 'x)))))
  (remove-if-not (lambda (square) (eq? (first (square-name square)) 'board))
		 *squares*))
 (update-staging-area! (squares->board)))



(define (reconstruct-board-new! pnm camera)
 (unless (every square-point
		(remove-if (lambda (square)
			    (not (equal? (first (square-name square)) 'board)))
			   *squares*))
  (message "You must first click on Find Board")
  (abort))
 (set! *ellipses-to-display* (make-vector *number-of-cameras* '()))
 (set! *board-squares-to-display* (make-vector *number-of-cameras* '()))
 (set! *board-os-to-display* (make-vector *number-of-cameras* '()))
 (set! *lines-to-display* (make-vector *number-of-cameras* '()))
 (let ((o-points (invariants pnm
			     (- *o-ellipse-finder-threshold* 10) ;;range from (threshold - 10)
			     (+ *o-ellipse-finder-threshold* 10) ;;      to   (threshold + 10)
			     5                                   ;;step-size
			     '()))
       (board-square-points (pnm->board-square-ellipses pnm)))
  (for-each
   (lambda (square)
    (set-square-contents!
     square
     (cond
      ((and (not (some (lambda (o-point)
		    (piece-in-range o-point (square-point square)))
		   o-points))
	    (not (some (lambda (board-square-point)
		    (empty-square-in-range board-square-point (square-point square)))
		   board-square-points)))
       'x)
      ((some (lambda (o-point)
		    (piece-in-range o-point (square-point square)))
		   o-points)
       'o)
      (else #f))))
   (remove-if-not (lambda (square) (eq? (first (square-name square)) 'board))
		  *squares*)))
 (update-staging-area! (squares->board)))



;;; Game Independent

(define (squares->board)
 (define (get-square-contents i j)
  (case (square-contents (board-square-at i j))
   ((x) 1)
   ((o) -1)
   ((#f) 0)
   (else (fuck-up))))
 (map-n (lambda (i) (map-n (lambda (j) (get-square-contents i j)) 3)) 3))

(define (achieve! board)
 (define (at i j)
  (case (list-ref (list-ref (second board) i) j)
   ((1) 'x)
   ((-1) 'o)
   ((0) #f)))
 (define (find-x)
  (minp (lambda (square1 square2)
	 (or (eq? (first (square-name square1)) 'board)
	     (and (not (eq? (first (square-name square2)) 'board))
		  (< (third (square-name square1))
		     (third (square-name square2))))))
	(remove-if-not
	 (lambda (square)
	  (and (eq? (square-contents square) 'x)
	       (or (and (eq? (first (square-name square)) 'board)
			(not (eq? (at (second (square-name square))
				      (third (square-name square)))
				  'x)))
		   (eq? (first (square-name square)) 'staging))))
	 *squares*)))
 (define (find-o)
  (minp (lambda (square1 square2)
	 (or (eq? (first (square-name square1)) 'board)
	     (and (not (eq? (first (square-name square2)) 'board))
		  (< (third (square-name square1))
		     (third (square-name square2))))))
	(remove-if-not
	 (lambda (square)
	  (and (eq? (square-contents square) 'o)
	       (or (and (eq? (first (square-name square)) 'board)
			(not (eq? (at (second (square-name square))
				      (third (square-name square)))
				  'o)))
		   (eq? (first (square-name square)) 'staging))))
	 *squares*)))
 (define (find-board-space piece)
   (call-with-current-continuation
    (lambda (k)
      (for-each-n
       (lambda (i)
	 (for-each-n
	  (lambda (j)
	    (when
		(and (= (list-2d-ref (second board) i j) piece)
		     (zero? (get-square-contents i j)))
	      (k (board-square-at i j))))
	  3))
       3))))
 (define (find-future-board-space piece)
   (call-with-current-continuation
    (lambda (k)
      (for-each-n
       (lambda (i)
	 (for-each-n
	  (lambda (j)
	    (when
		(and (= (list-2d-ref (second board) i j) piece)
		     (not (= (get-square-contents i j) piece)))
	      (k (board-square-at i j))))
	  3))
       3))))
 (define (find-space-for-x)
   (if (find-board-space 1)
       (find-board-space 1)
       (minp (lambda (square1 square2)
	       (> (third (square-name square1)) (third (square-name square2))))
	     (remove-if-not (lambda (square)
			      (and (not (square-contents square))
				   (eq? (first (square-name square)) 'staging)
				   (eq? (second (square-name square)) 'left)))
			    *squares*))))
 (define (get-square-contents i j)
  (case (square-contents (board-square-at i j))
   ((x) 1)
   ((o) -1)
   ((#f) 0)
   (else (fuck-up))))
 (define (find-space-for-o)
  (if
   (find-board-space -1)
   (find-board-space -1)
   (minp (lambda (square1 square2)
	   (> (third (square-name square1)) (third (square-name square2))))
	 (remove-if-not (lambda (square)
			  (and (not (square-contents square))
			       (eq? (first (square-name square)) 'staging)
			       (eq? (second (square-name square)) 'right)))
			*squares*))))
 (define (reconstruct-if-necessary!)
  (when *reconstruct?*
   (for-each-n (lambda (camera)
		(hack-track-page-in-frames camera 0 1)
		(hack-track-capture-frames camera 0 1 1))
	       *number-of-cameras*)
   (reconstruct-board-new! (hack-track-frame->ppm *camera* 0) *camera*)
   (redraw-display-pane)
   (set! *end* 0)))
 (define (perform-action i j)
  (cond ((and (eq? (at i j) 'x)
	      (eq? (square-contents (board-square-at i j)) #f))
	 (pick-up (find-x))
	 (put-down (board-square-at i j))
	 (reconstruct-if-necessary!))
	((and (eq? (at i j) 'x)
	      (eq? (square-contents (board-square-at i j)) 'o))
	 (pick-up (board-square-at i j))
	 (put-down (find-space-for-o))
	 (reconstruct-if-necessary!)
	 (pick-up (find-x))
	 (put-down (board-square-at i j))
	 (reconstruct-if-necessary!))
	((and (eq? (at i j) 'o)
	      (eq? (square-contents (board-square-at i j)) #f))
	 (pick-up (find-o))
	 (put-down (board-square-at i j))
	 (reconstruct-if-necessary!))
	((and (eq? (at i j) 'o)
	      (eq? (square-contents (board-square-at i j)) 'x))
	 (pick-up (board-square-at i j))
	 (put-down (find-space-for-x))
	 (reconstruct-if-necessary!)
	 (pick-up (find-o))
	 (put-down (board-square-at i j))
	 (reconstruct-if-necessary!))
	((and (eq? (at i j) #f)
	      (eq? (square-contents (board-square-at i j)) 'x))
	 (let ((future (find-future-board-space 1)))
	   (if future
	       (begin (perform-action (second (square-name future))
				      (third (square-name future)))
		      (perform-action i j))
	       (begin
		 (pick-up (board-square-at i j))
		 (put-down (find-space-for-x))
		 (reconstruct-if-necessary!)))))
	((and (eq? (at i j) #f)
	      (eq? (square-contents (board-square-at i j)) 'o))
	 (let ((future (find-future-board-space -1)))
	   (if future
	       (begin (perform-action (second (square-name future))
				      (third (square-name future)))
		      (perform-action i j))
	       (begin
		 (pick-up (board-square-at i j))
		 (put-down (find-space-for-o))
		 (reconstruct-if-necessary!)))))))

 (for-each-n (lambda (i) (for-each-n
			  (lambda (j)
			    (perform-action i j))
			  3))
	     3)
 (unless (equal? (squares->board) (second board))
  (achieve! board)))

(define (player board) (first board))

(define (initial-board)
 (case *game*
  ((tic-tac-toe) (tic-tac-toe-initial-board))
  ((hexapawn) (hexapawn-initial-board))
  ((prolog) (prolog-initial-board))
  (else (fuck-up))))

(define (moves board)
 (case *game*
  ((tic-tac-toe) (tic-tac-toe-moves board))
  ((hexapawn) (hexapawn-moves board))
  ((prolog) (prolog-moves board))
  (else (fuck-up))))

(define (win0 board)
 (case *game*
  ((tic-tac-toe) (tic-tac-toe-win0 board))
  ((hexapawn) (hexapawn-win0 board))
  ((prolog) (prolog-win0 board))
  (else (fuck-up))))

(define (maximize f l limit)
 (define (loop best-so-far l)
  (cond ((>= best-so-far limit) 1)
	((null? l) best-so-far)
	(else (loop (max (f (first l) best-so-far) best-so-far) (rest l)))))
 (loop -1 l))

(define (win*-alpha/beta board limit)
 (let* ((moves-board (moves board))
	(player-board (player board))
	(win-board (win0 board)))
 (cond ((not (zero? win-board)) win-board)
       ((null? moves-board) 0)
       (else
	(* (player board)
	   (maximize (lambda (move limit)
		      (* player-board
			 (win*-alpha/beta move (* player-board limit))))
		     moves-board
		     (* player-board limit)))))))

(define (win* board) (win*-alpha/beta board (player board)))

(define (optimal-moves board)
 (let ((w* (* (player board) (win* board))))
  (remove-if-not (lambda (move) (= (* (player board) (win* move)) w*))
		 (moves board))))

(define (play! player)
 (let ((board (list player (squares->board))))
  (when (or (not (zero? (win0 board))) (null? (moves board)))
   (message "Game over") (abort))
  (achieve! (random-member ((if *optimal?* optimal-moves moves) board)))))

(define (play-game!)
 (hack-track-unblock)
 (hack-track-set-interrupts?! #t)
 (achieve! (initial-board))
 (let loop ((player (player (initial-board))) (boards '()))
  (interruptible-usleep 100000)
  (let* ((board (list player (squares->board))) (outcome (win0 board)))
   (cond ((zero? outcome)
	  (cond ((null? (moves board))
		 (message "Draw")
		 (interruptible-usleep 1000000)
		 (list (reverse (cons board boards)) outcome))
		(else (play! player)
		      (loop (- player) (cons board boards)))))
	 ((= outcome 1)
	  (message "X wins")
	  (interruptible-usleep 1000000)
	  (list (reverse (cons board boards)) outcome))
	 ((= outcome -1)
	  (message "O wins")
	  (interruptible-usleep 1000000)
	  (list (reverse (cons board boards)) outcome))
	 (else (fuck-up))))))

;;; Tic Tac Toe

(define (tic-tac-toe-initial-board) '(1 ((0 0 0) (0 0 0) (0 0 0))))

(define (tic-tac-toe-moves board)
 (define (discard-non-empty row)
  (cond ((null? row) '())
	((zero? (third (first row)))
	 (cons (first row) (discard-non-empty (rest row))))
	(else (discard-non-empty (rest row)))))
 (define (list-replace-ith list i x)
  (if (zero? i)
      (cons x (rest list))
      (cons (first list) (list-replace-ith (rest list) (- i 1) x))))
 (map (lambda (move)
       (list (- (first board))
	     (list-replace-ith
	      (second board)
	      (first move)
	      (list-replace-ith (list-ref (second board) (first move))
				(second move)
				(player board)))))
      (reduce append
	      (map-indexed
	       (lambda (row i)
		(discard-non-empty
		 (map-indexed (lambda (element j) (list i j element)) row)))
	       (second board))
	      '())))

(define (tic-tac-toe-win0 board)
 (define (at board i j) (list-ref (list-ref (second board) i) j))
 (define (win? player)
  (let ((n (length (second board))))
   (or
    (some-n (lambda (i) (every-n (lambda (j) (= (at board i j) player)) n)) n)
    (some-n (lambda (j) (every-n (lambda (i) (= (at board i j) player)) n)) n)
    (every-n (lambda (i) (= (at board i i) player)) n)
    (every-n (lambda (i) (= (at board i (- (- n 1) i)) player)) n))))
 (cond ((win? 1) 1)
       ((win? -1) -1)
       (else 0)))

;;; Hexapawn

(define (hexapawn-initial-board) '(1 ((1 1 1) (0 0 0) (-1 -1 -1))))

(define (list-2d-replace list i j x)
  (list-replace list
		i
		(list-replace (list-ref list i) j x)))

(define (list-2d-ref list i j) (list-ref (list-ref list i) j))

(define (hexapawn-board-move board from-i from-j to-i to-j)
  (list-2d-replace
   (list-2d-replace board
		    to-i
		    to-j
		    (list-2d-ref board
				 from-i
				 from-j))
   from-i
   from-j
   0))

(define (hexapawn-moves board)
  (define (hexapawn-piece-at i j board) (list-2d-ref board i j))
  (define (hexapawn-piece-at? i j board) (not (zero? (list-2d-ref board i j))))
  (define (hexapawn-on-board? i board)
    (and (>= i 0)
	 (< i (length board))))
  (define (hexapawn-can-capture? piece direction i j board)
    (and
     (hexapawn-on-board? (+ i piece) board)
     (hexapawn-on-board? (+ j direction) board)
     (hexapawn-piece-at? (+ i piece) (+ j direction) board)
     (not (eq?
	   (hexapawn-piece-at (+ i piece) (+ j direction) board)
	   piece))))
  (define (hexapawn-can-capture-l? piece i j board)
    (hexapawn-can-capture? piece 1 i j board))
  (define (hexapawn-can-capture-r? piece i j board)
    (hexapawn-can-capture? piece -1 i j board))
  (define (hexapawn-move-forward? piece i j board)
    (and
     (hexapawn-on-board? (+ i piece) board)
     (not (hexapawn-piece-at? (+ i piece) j board))))
  (define (hexapawn-move-backward? piece i j pieces)
    (and
     (hexapawn-on-board? (- i piece) pieces)
     (not (hexapawn-piece-at? (- i piece) j pieces))))
  (define (hexapawn-move-left? piece i j pieces)
    (and
     (hexapawn-on-board? (- j 1) pieces)
     (not (hexapawn-piece-at? i (- j 1) pieces))))
  (define (hexapawn-move-right? piece i j pieces)
    (and
     (hexapawn-on-board? (+ j 1) pieces)
     (not (hexapawn-piece-at? i (+ j 1) pieces))))
  (define (hexapawn-move-sideways? piece direction i j pieces)
    (and
     (hexapawn-on-board? (+ i piece) pieces)
     (hexapawn-on-board? (+ j direction) pieces)
     (not (hexapawn-piece-at? (+ i piece) (+ j direction) pieces))))
  (define (hexapawn-move-back-sideways? piece direction i j pieces)
    (and
     (hexapawn-on-board? (- i piece) pieces)
     (hexapawn-on-board? (+ j direction) pieces)
     (not (hexapawn-piece-at? (- i piece) (+ j direction) pieces))))

  (let* ((player (first board)) (board (second board)) (n (length board)))
    (map
     (lambda (x)
       `(,(- player) ,x))
     (reduce
      append
      (map-indexed
       (lambda (row i)
	 (reduce
	  append
	  (map-indexed
	   (lambda (piece j)
	     (case *game-variant*
	       ((regular)
		(if (= piece player)
		    (append
		     (if (hexapawn-move-forward? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) j))
			 '())
		     (if (hexapawn-can-capture-l? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (+ j 1)))
			 '())
		     (if (hexapawn-can-capture-r? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (- j 1)))
			 '()))
		    '()))
	       ((variant-a)
		(if (= piece player)
		    (append
		     (if (hexapawn-move-sideways? piece 1 i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (+ j 1)))
			 '())
		     (if (hexapawn-move-sideways? piece -1 i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (- j 1)))
			 '())
		     (if (hexapawn-can-capture-l? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (+ j 1)))
			 '())
		     (if (hexapawn-can-capture-r? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (- j 1)))
			 '()))
		    '()))
	       ((variant-b)
		(if (= piece player)
		    (append
		     (if (hexapawn-move-sideways? piece 1 i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (+ j 1)))
			 '())
		     (if (hexapawn-move-sideways? piece -1 i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (- j 1)))
			 '())
		     (if (hexapawn-move-back-sideways? piece 1 i j board)
			 `(,(hexapawn-board-move board i j (- i piece) (+ j 1)))
			 '())
		     (if (hexapawn-move-back-sideways? piece -1 i j board)
			 `(,(hexapawn-board-move board i j (- i piece) (- j 1)))
			 '())
		     (if (hexapawn-can-capture-l? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (+ j 1)))
			 '())
		     (if (hexapawn-can-capture-r? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (- j 1)))
			 '()))
		    '()))
	       ((variant-c)
		(if (= piece player)
		    (append
		     (if (hexapawn-move-forward? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) j))
			 '())
		     (if (hexapawn-move-backward? piece i j board)
			 `(,(hexapawn-board-move board i j (- i piece) j))
			 '())
		     (if (hexapawn-can-capture-l? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (+ j 1)))
			 '())
		     (if (hexapawn-can-capture-r? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (- j 1)))
			 '()))
		    '()))
	       ((variant-d)
		(if (= piece player)
		    (append
		     (if (hexapawn-move-forward? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) j))
			 '())
		     (if (hexapawn-move-backward? piece i j board)
			 `(,(hexapawn-board-move board i j (- i piece) j))
			 '())
		     (if (hexapawn-move-left? piece i j board)
			 `(,(hexapawn-board-move board i j i (- j 1)))
			 '())
		     (if (hexapawn-move-right? piece i j board)
			 `(,(hexapawn-board-move board i j i (+ j 1)))
			 '())
		     (if (hexapawn-can-capture-l? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (+ j 1)))
			 '())
		     (if (hexapawn-can-capture-r? piece i j board)
			 `(,(hexapawn-board-move board i j (+ i piece) (- j 1)))
			 '()))
		    '()))))
	       row)
	     '()))
	  board)
	 '()))))

(define (hexapawn-win0 board)
 (cond ((some (lambda (current) (= current 1)) (last (second board))) 1)
       ((some (lambda (current) (= current -1)) (first (second board))) -1)
       ((null? (hexapawn-moves board)) (- (first board)))
       (else 0)))

;;; Prolog

(define (prolog-board->scheme-board board)
 (map (lambda (row)
       (map (lambda (square)
	     (case square
	      ((none) 0)
	      ((x) 1)
	      ((o) -1)))
	    row))
      board))

(define (prolog-player->scheme-player player)
 (case player
  ((player_x) 1)
  ((player_o) -1)
  (else (fuck-up))))

(define (prolog-outcome->scheme-outcome outcome)
 (case outcome
  ((x_wins) 1)
  ((o_wins) -1)
  ((neither) 0)
  (else (fuck-up))))

(define (scheme-string-list->prolog-string l)
 (string-append "[" (reduce (lambda (x y) (string-append x "," y)) l "") "]"))

(define (scheme-board->prolog-board board)
 (scheme-string-list->prolog-string
  (map (lambda (row)
	(scheme-string-list->prolog-string
	 (map (lambda (square)
	       (case square
		((0) "none")
		((1) "x")
		((-1) "o")))
	      row)))
       board)))

(define (scheme-player->prolog-player player)
 (case player
  ((1) "player_x")
  ((-1) "player_o")
  (else (fuck-up))))

(define (scheme-outcome->prolog-outcome outcome)
 (case outcome
  ((1) "x_wins")
  ((-1) "o_wins")
  ((0) "draw")
  (else (fuck-up))))

(define (prolog-initial-board)
 (let ((board
	(prolog-command
	 (format #f "initial_board(Board,Player),writeln([Player,Board])."))))
  (list (prolog-player->scheme-player (first board))
	(prolog-board->scheme-board (second board)))))

(define (prolog-moves board)
 (map (lambda (board1)
       (list (- (first board)) (prolog-board->scheme-board board1)))
      (prolog-command
       (format #f
	       "findall(Board,legal_move(~a,~a,Board),Board),writeln(Board)."
	       (scheme-player->prolog-player (first board))
	       (scheme-board->prolog-board (second board))))))

(define (prolog-win0 board)
 (prolog-outcome->scheme-outcome
  (prolog-command
   (format
    #f "((outcome(~a,~a,Outcome),writeln(Outcome),!);writeln('neither'))."
    (scheme-player->prolog-player (first board))
    (scheme-board->prolog-board (second board))))))

;;; Progol

(define (write-game boards outcome port)
 (format port "initial_board(~a,~a).~%"
	 (scheme-board->prolog-board (second (first boards)))
	 (scheme-player->prolog-player (first (first boards))))
 (for-each (lambda (board1 board2)
	     (format port ":-outcome(~a,~a,x_wins).~%"
		    (scheme-player->prolog-player (first board1))
		    (scheme-board->prolog-board (second board1)))
	    (format port ":-outcome(~a,~a,o_wins).~%"
		    (scheme-player->prolog-player (first board1))
		    (scheme-board->prolog-board (second board1)))
	    (format port "legal_move(~a,~a,~a).~%"
		    (scheme-player->prolog-player (first board1))
		    (scheme-board->prolog-board (second board1))
		    (scheme-board->prolog-board (second board2))))
	   (but-last boards)
	   (rest boards))
 (format port "outcome(~a,~a,~a).~%"
	 (scheme-player->prolog-player (first (last boards)))
	 (scheme-board->prolog-board (second (last boards)))
	 (scheme-outcome->prolog-outcome outcome))
 (format port ":- outcome(~a,~a,~a).~%"
	 (scheme-player->prolog-player (first (last boards)))
	 (scheme-board->prolog-board (second (last boards)))
	 (scheme-outcome->prolog-outcome (- outcome))))

(define (compose-background-knowledge pathname)
  (append (read-file pathname)
	  (read-file (format #f "~a/imitate/tool1/background-knowledge.pl"
			     (getenv "HOME")))))

(define (learn-predicate background-knowledge-pathname games correct-cover? type)
 (let ((input-pathname (format #f "/tmp/progol-input-~a-~a-~a.pl"
			       (getenv "USERNAME") (getpid) type))
       (output-pathname (format #f "/tmp/progol-output-~a-~a-~a.pl"
				(getenv "USERNAME") (getpid) type))
       (background-knowledge
	(compose-background-knowledge background-knowledge-pathname))
       (game-data
	(begin
	  (let ((string (open-output-string)))
	  (for-each (lambda (game)
		      (write-game (first game) (second game) string))
		    games)
	  (get-output-string string)))))
  (rm input-pathname)
  (rm output-pathname)
  (call-with-output-file input-pathname
   (lambda (port)
     (for-each (lambda (line) (display line port) (newline port)) background-knowledge)
     (format port game-data)))

  (if correct-cover?
      (system (format #f "~~/imitate/progol/progol ~a | perl -0777 -ne 'print \"$2\\n\" while /Result of search([^\\]]*).(.*?)\\n\\[/gs' > ~a"
		      input-pathname output-pathname))
      (system (format #f "~~/imitate/progol/progol ~a >~a"
		      input-pathname output-pathname)))
  (let ((lines (read-file output-pathname)))
    (rm input-pathname)
    (rm output-pathname)
    (if correct-cover?
	(begin
	  (reduce
	   string-append
	   (find-best-rules
	    (compose-background-knowledge background-knowledge-pathname)
	    (extract-clauses type lines)
	    game-data)
	   ""))
	(let ((reverse-lines (rest (rest (rest (reverse lines))))))
	  (reverse (sublist reverse-lines 0
			    (- (position-if (lambda (x) (prefix? "[" x))
					    reverse-lines) 1))))))))

(define (untabify s)
 ;; This doesn't do the right thing since it doesn't know about tab stops.
 ;; It just replaces tabs with 8 spaces.
 (reduce
  string-append
  (map (lambda (char) (if (char=? char #\tab) "        " (string char)))
       (string->list s))
  ""))

(define (learn-rules! k l m)
 (let ((games '()))
  (let loop ()
   (message
    (format
     #f
     "Playing game ~a, having ~a non-draws, out of at least ~a non-drawing games"
     (length games)
     (count-if (lambda (game) (not (zero? (second game)))) games)
     m))
   (when *optimal?*
    (message "You really don't want to do this with optimal play")
    (abort))
   (set! games (cons (play-game!) games))
   (unless (and
	    (>= (length games) k)
	    (>= (length games) l)
	    (>= (count-if (lambda (game) (not (zero? (second game)))) games)
		m))
    (loop)))
  (message "")
  (achieve! (initial-board))
  (park-arm)
  (message "Learning")
  
  (let* ((initial-board
	  (learn-predicate
	   (format #f "~a/imitate/tool1/background-knowledge-initial-board.pl"
		   (getenv "HOME"))
	   (sublist games 0 k)
	   #f
	   "initial"))
	 (moves
	  (learn-predicate
	   (format #f "~a/imitate/tool1/background-knowledge-legal-move.pl"
		   (getenv "HOME"))
	   (sublist games 0 l)
	   #f
	   "legal_move"))
	 (temp-background-knowledge-outcome
	  (format #f "/tmp/prolog-background-knowledge-outcome-predicate-~a-~a.pl"
		  (getenv "USERNAME") (getpid)))
	 (outcome
	  (learn-predicate
	   (begin
	     (write-file
	      (append
	       (read-file
		(format #f "~a/imitate/tool1/background-knowledge-outcome.pl"
			(getenv "HOME")))
	       moves)
	      temp-background-knowledge-outcome)
	     temp-background-knowledge-outcome)
	   (sublist (remove-if (lambda (game) (zero? (second game))) games)
		    0
		    m)
	   #t
	   "outcome")))
    (set! *rules* (map untabify (append initial-board moves outcome))))

  (write-file
   (append
    (read-file (format #f "~a/imitate/tool1/background-knowledge.pl"
		       (getenv "HOME")))
    (read-file (format #f "~a/imitate/tool1/prolog-play-predicates.pl"
		       (getenv "HOME")))
    *rules*)
   (format #f "~a/imitate/tool1/learned-rules.pl" (getenv "HOME")))
  (redraw-display-pane)
  (message "")))

;;; Find a subset of the rules that explains our data

(define (progol-test-line->numbers string)
  (let* ((line (string->list string))
	 (body (sublist line (+ 1 (position #\| line)) (length line)))
	 (number1 (string->number
		   (list->string (sublist body 0 (position #\| body)))))
	 (body2 (sublist body (+ 1 (position #\| body)) (length body)))
	 (number2 (string->number
		   (list->string (sublist body2 0 (position #\| body2))))))
    `(,number1 ,number2)))

(define (tested-rules-good? output)
  (and (zero? (second (first output)))
       (zero? (first (second output)))))

(define (test-rules background rules data)
  (let ((b-file (unique-temporary-file "/tmp/b.pl"))
	(r-file (unique-temporary-file "/tmp/r.pl"))
	(d-file (unique-temporary-file "/tmp/d.pl"))
	(run-file (unique-temporary-file "/tmp/run.pl"))
	(out-file (unique-temporary-file "/tmp/out.pl")))
    (write-file background b-file)
    (write-file rules r-file)
    (write-file data d-file)
    (write-file
     `(,(format #f
		"['~a']?~%['~a']?~%test('~a')?~%" b-file r-file d-file))
     run-file)
    (system (format #f "progol < ~a > ~a" run-file out-file))
    (let ((output (read-file out-file)))
      (rm out-file)
      (rm b-file)
      (rm r-file)
      (rm d-file)
      (map
       progol-test-line->numbers
       (remove-if-not (lambda (x) (substring? "P|" x)) output)))))

(define (all-ordered-subsets rules)
  (if (null? rules)
      '()
      (remove-duplicates
       (cons
	rules
	(reduce
	 append
	 (map (lambda (x) (all-ordered-subsets (remove x rules))) rules)
	 '())))))

(define (find-best-rules background rules data)
  (call-with-current-continuation
   (lambda (k)
     (for-each
      (lambda (x)
	(if (tested-rules-good? (test-rules background x data))
	    (k x)))
      (sort (all-ordered-subsets rules) < length)
     (k rules)))))

;;; Screen recording

(define (window-id name)
 (let ((pathname
	(format #f "/tmp/xwininfo-~a-~a.text" (getenv "USERNAME") (getpid))))
  (rm pathname)
  (system
   (format #f "xwininfo -name ~a|fgrep id:|cut -d ' ' -f 4 >~a" name pathname))
  (let ((id (first (read-file pathname))))
   (rm pathname)
   id)))

(define (start-recording! fps)
 (system
  (format
   #f
   "recordmydesktop &>/dev/null -windowid ~a --no-sound --on-the-fly-encoding -fps ~a&"
   (window-id *program*)
   fps)))

(define (stop-recording!) (system "killall &>/dev/null recordmydesktop"))

;;; Commands

(define (viewfinder-command)
 (message "")
 (set! *mode* 'viewfinder)
 (redraw-display-pane)
 (redraw-buttons)
 (for-each-n (lambda (camera) (hack-track-turn-on-viewfinder! camera 1))
	     *number-of-cameras*))

(define (overlay-command)
 (message "")
 (unless *end*
  (message "You must first take a snapshot or video")
  (abort))
 (for-each-n (lambda (camera) (hack-track-turn-off! camera))
	     *number-of-cameras*)
 (set! *mode* 'overlay)
 (set! *t* 0)
 (redraw-display-pane)
 (redraw-buttons))

(define (beginning-command)
 (message "")
 (unless (eq? *mode* 'overlay)
  (message "Can only do this in overlay mode")
  (abort))
 (unless (zero? *t*)
  (set! *t* 0)
  (redraw-buttons)
  (redraw-display-pane)))

(define (decrement-t-command)
 (message "")
 (unless (eq? *mode* 'overlay)
  (message "Can only do this in overlay mode")
  (abort))
 (when (zero? *t*) (message "At beginning of movie") (abort))
 (set! *t* (- *t* 1))
 (redraw-buttons)
 (redraw-display-pane))

(define (increment-t-command)
 (message "")
 (unless (eq? *mode* 'overlay)
  (message "Can only do this in overlay mode")
  (abort))
 (when (= *t* *end*) (message "At end of movie") (abort))
 (set! *t* (+ *t* 1))
 (redraw-buttons)
 (redraw-display-pane))

(define (end-command)
 (message "")
 (unless (eq? *mode* 'overlay)
  (message "Can only do this in overlay mode")
  (abort))
 (unless (= *t* *end*)
  (set! *t* *end*)
  (redraw-buttons)
  (redraw-display-pane)))

;;; Top Level

(define-application viewer *width* *height* 0 9 8
 (lambda ()
  (set! *width* 320)
  (set! *height* 240)
  (set! *mode* 'viewfinder)
  (set! *end* #f)
  (set! *lines* '())
  (set! *rules* '())
  (set! *first-line* 0)
  (set! abort-command
	(lambda ()
	 (message "")
	 (case *mode*
	  ((viewfinder)
	   (for-each-n
	    (lambda (camera) (hack-track-turn-on-viewfinder! camera 1))
	    *number-of-cameras*))
	  ((overlay)
	   (for-each-n (lambda (camera) (hack-track-turn-off! camera))
		       *number-of-cameras*))
	  (else (fuck-up)))
	 (abort)))
  (set! *ellipses-to-display* (make-vector *number-of-cameras* '()))
  (set! *board-squares-to-display* (make-vector *number-of-cameras* '()))
  (set! *board-os-to-display* (make-vector *number-of-cameras* '()))
  (set! *lines-to-display* (make-vector *number-of-cameras* '()))
  (cd (format #f "~a/imitate/stuff" (getenv "HOME")))
  ;; First row
  (define-button 0 0 "Help" #f
   (lambda ()
    (message "")
    (for-each-n (lambda (camera) (hack-track-turn-off! camera))
		*number-of-cameras*)
    (help-command)))
  (define-button 1 0 "Abort" #f abort-command)
  (define-button 2 0 "Snapshot" #f
   (lambda ()
    (message "")
    (for-each-n (lambda (camera) (hack-track-turn-on-capturing! camera 0 1 1))
		*number-of-cameras*)
    (wait-for-next-frame 0)
    (set! *ellipses-to-display*
	  (map-n-vector
	   (lambda (camera)
	    (pnm->ellipses (ppm->pgm (hack-track-frame->ppm camera 0))
			   *board-square-ellipse-finder-threshold*))
	   *number-of-cameras*))
    (set! *board-squares-to-display*
	  (map-n-vector
	   (lambda (camera)
	    (pnm->board-square-ellipses
	     (ppm->pgm (hack-track-frame->ppm camera 0))))
	   *number-of-cameras*))
    (set! *board-os-to-display*
	  (map-n-vector
	   (lambda (camera)
	    (pnm->ellipses-of-area (ppm->pgm (hack-track-frame->ppm camera 0))
				   *o-ellipse-finder-threshold*
				   *o-ellipse-finder-min-area*
				   *o-ellipse-finder-max-area*))
	   *number-of-cameras*))
    (set! *lines-to-display*
	  (map-n-vector (lambda (camera)
			 (remove-if
			  (lambda (x) (< (line-segment-length x) 40))
			  (pnm->line-segments (hack-track-frame->ppm camera 0)
					      1.0 0.3 0.3 12 15 10 5)))
			*number-of-cameras*))
    (set! *end* 0)
    (overlay-command)))
  (define-button 3 0 "Video" #f
   (lambda ()
    (message "")
    (for-each-n (lambda (camera) (hack-track-turn-on-capturing! camera 0 1 20))
		*number-of-cameras*)
    (interruptible-usleep 4000000)
    (set! *end* 19)
    (viewfinder-command)))
  (define-button 4 0 "Write PPM" (lambda () (eq? *mode* 'overlay))
   (lambda ()
    (message "")
    (unless (eq? *mode* 'overlay)
     (message "Can only do this in overlay mode")
     (abort))
    (for-each-n (lambda (camera)
		 (write-pnm (hack-track-frame->ppm camera *t*)
			    (format #f "/tmp/snapshot~s.ppm" camera)))
		*number-of-cameras*)))
  (define-button 5 0 "Play" (lambda () *end*)
   (lambda ()
    (message "")
    (unless *end*
     (message "You must first take a snapshot or video")
     (abort))
    (for-each-n (lambda (t)
		 (show-frame t)
		 (xflush *display*)
		 (usleep (inexact->exact (round (/ 1000000.0 *frame-rate*)))))
		(+ *end* 1))
    (redraw-display-pane)))
  (define-toggle-button 6 0 "Optimal" *optimal?* (lambda () (message "")))
  (define-button 7 0 "Quit" #f quit)
  ;; Second row
  (define-toggle-button 0 1 "Record" *record?*
   (lambda ()
    (message "")
    (if *record?* (start-recording! 10) (stop-recording!))))
  (define-cycle-button 1 1 *game-variant*
   (lambda () (message ""))
   (regular "Regular")
   (variant-a "Variant A")
   (variant-b "Variant B")
   (variant-c "Variant C")
   (variant-d "Variant D"))
  (define-button 2 1 "Viewfinder" (lambda () (eq? *mode* 'viewfinder))
   viewfinder-command)
  (define-button 3 1 "Overlay" (lambda () (eq? *mode* 'overlay))
   overlay-command)
  (define-button 4 1 "Beginning" (lambda () (eq? *mode* 'overlay))
   beginning-command)
  (define-button 5 1
   (lambda () (if (eq? *mode* 'overlay) (format #f "-T ~s" *t*) "-T"))
   (lambda () (eq? *mode* 'overlay))
   decrement-t-command)
  (define-button 6 1
   (lambda () (if (eq? *mode* 'overlay) (format #f "+T ~s" *t*) "+T"))
   (lambda () (eq? *mode* 'overlay))
   increment-t-command)
  (define-button 7 1 "End" (lambda () (eq? *mode* 'overlay)) end-command)
  ;; Third row
  (define-button 0 2 "-Brightness" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-brightness!
     *camera* (max (- (v4l-get-brightness *camera*) 512) 0))
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 1 2 "+Brightness" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-brightness!
     *camera* (min (+ (v4l-get-brightness *camera*) 512) 65535))
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 2 2 "-Contrast" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-contrast! *camera* (max (- (v4l-get-contrast *camera*) 1024) 0))
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 3 2 "+Contrast" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-contrast!
     *camera* (min (+ (v4l-get-contrast *camera*) 1024) 65535))
    (recache-strings!)
    (redraw-display-pane)))
  ;; Hue doesn't work on our cameras.
  (define-button 4 2 "-Colour" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-colour! *camera* (min (- (v4l-get-colour *camera*) 327) 0))
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 5 2 "+Colour" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-colour! *camera* (max (+ (v4l-get-colour *camera*) 327) 65535))
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 6 2 "-Whiteness" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-whiteness! *camera* (min (- (v4l-get-whiteness *camera*) 2048) 0))
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 7 2 "+Whiteness" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-whiteness!
     *camera* (max (+ (v4l-get-whiteness *camera*) 2048) 65535))
    (recache-strings!)
    (redraw-display-pane)))
  ;; Fourth row
  (define-button 0 3 "Home" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-home! *camera* 1 1)
    (wait-for-pan-tilt *camera*)
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 1 3 "Left" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-pan-tilt! *camera* -250 0 0)
    (wait-for-pan-tilt *camera*)
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 2 3 "Right" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-pan-tilt! *camera* 250 0 0)
    (wait-for-pan-tilt *camera*)
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 3 3 "Up" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-pan-tilt! *camera* 0 250 0)
    (wait-for-pan-tilt *camera*)
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 4 3 "Down" #f
   (lambda ()
    (message "")
    (when (eq? *mode* 'overlay)
     (message "Can't do this in overlay mode")
     (abort))
    (v4l-set-pan-tilt! *camera* 0 -250 0)
    (wait-for-pan-tilt *camera*)
    (recache-strings!)
    (redraw-display-pane)))
  (define-button 5 3 "Load Vision" #f
   (lambda ()
    (message "")
    (load-ellipse-finder-parameters!)))
  (define-toggle-button 6 3 "Vision" *reconstruct?* (lambda () (message "")))
  (define-button 7 3 "Learn Rules" #f
   (lambda ()
    (message "")
    (if (and (= (number-of-fields *pathname*) 3)
	     (string->number (field-ref *pathname* 0))
	     (string->number (field-ref *pathname* 1))
	     (string->number (field-ref *pathname* 2)))
	(learn-rules!
	 (inexact->exact (string->number (field-ref *pathname* 0)))
	 (inexact->exact (string->number (field-ref *pathname* 1)))
	 (inexact->exact (string->number (field-ref *pathname* 2))))
	(learn-rules! 1 2 4))))
  ;; Fifth row
  (define-button 0 4 "Park Arm" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (park-arm)
    (redraw-display-pane)))
  (define-button 1 4 "Reset Arm" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (reset-arm)
    (redraw-display-pane)))
  (define-button 2 4 "Relax Arm" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (relax-shoulder-and-elbow)
    (redraw-display-pane)))
  (define-button 3 4 "Center Head" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (center-head)
    (redraw-display-pane)))
  (define-button 4 4 "Relax Head" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (relax-head)
    (redraw-display-pane)))
  (define-button 5 4 "Edit Square" #f
   (lambda ()
    (message "")
    (unless *square* (message "No square selected") (abort))
    (case (first (square-name *square*))
     ((board)
      (unless (and (= (number-of-fields *pathname*) 2)
		   (string->number (field-ref *pathname* 0))
		   (string->number (field-ref *pathname* 1)))
       (message "You need to supply values x and y")
       (abort))
      (matrix-set! *square-xs*
		   (second (square-name *square*))
		   (third (square-name *square*))
		   (string->number (field-ref *pathname* 0)))
      (matrix-set! *square-ys*
		   (second (square-name *square*))
		   (third (square-name *square*))
		   (string->number (field-ref *pathname* 1))))
     ((staging)
      (unless (and (= (number-of-fields *pathname*) 3)
		   (string->number (field-ref *pathname* 0))
		   (string->number (field-ref *pathname* 1))
		   (string->number (field-ref *pathname* 2)))
       (message "You need to supply values x, y, and z")
       (abort))
      (matrix-row-set! (case (second (square-name *square*))
			((left) *staging-area-left*)
			((right) *staging-area-right*)
			(else (fuck-up)))
		       (third (square-name *square*))
		       (vector (string->number (field-ref *pathname* 0))
			       (string->number (field-ref *pathname* 1))
			       (string->number (field-ref *pathname* 2)))))
     (else (fuck-up)))))
  (define-button 6 4 "Load Robot" #f
   (lambda ()
    (message "")
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (load-robot-parameters!)))
  (define-button 7 4 "Save Robot" #f
   (lambda ()
    (message "")
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (save-robot-parameters!)))
  ;; Sixth and Seventh rows
  (define-button 0 5 "Base Left" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-base (- *base* 10))
    (redraw-display-pane)))
  (define-button 0 6 "Base Right" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-base (+ *base* 10))
    (redraw-display-pane)))
  (define-button 1 5 "Shoulder Up" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-shoulder (- *shoulder* 10))
    (redraw-display-pane)))
  (define-button 1 6 "Shoulder Down" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-shoulder (+ *shoulder* 10))
    (redraw-display-pane)))
  (define-button 2 5 "Elbow Up" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-elbow (+ *elbow* 10))
    (redraw-display-pane)))
  (define-button 2 6 "Elbow Down" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-elbow (- *elbow* 10))
    (redraw-display-pane)))
  (define-button 3 5 "Hand Up" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-hand (+ *hand* 10))
    (redraw-display-pane)))
  (define-button 3 6 "Hand Down" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-hand (- *hand* 10))
    (redraw-display-pane)))
  (define-button 4 5 "LFinger Open" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-left-finger (- *left-finger* 10))
    (redraw-display-pane)))
  (define-button 4 6 "LFinger Close" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-left-finger (+ *left-finger* 10))
    (redraw-display-pane)))
  (define-button 5 5 "RFinger Open" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-right-finger (+ *right-finger* 10))
    (redraw-display-pane)))
  (define-button 5 6 "RFinger Close" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-right-finger (- *right-finger* 10))
    (redraw-display-pane)))
  (define-button 6 5 "Fingers Open" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-left-finger (- *left-finger* 10))
    (move-right-finger (+ *right-finger* 10))
    (redraw-display-pane)))
  (define-button 6 6 "Fingers Close" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-left-finger (+ *left-finger* 10))
    (move-right-finger (- *right-finger* 10))
    (redraw-display-pane)))
  (define-button 7 5 "Head Left" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-head (+ *head* 10))
    (redraw-display-pane)))
  (define-button 7 6 "Head Right" #f
   (lambda ()
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (move-head (- *head* 10))
    (redraw-display-pane)))
  ;; Eighth row
  (define-button 0 7 "Pick Up" #f
   (lambda ()
    (message "")
    (unless *square* (message "No square selected") (abort))
    (pick-up *square*)
    (set! *square* #f)))
  (define-button 1 7 "Put Down" #f
   (lambda ()
    (message "")
    (unless *square* (message "No square selected") (abort))
    (put-down *square*)
    (set! *square* #f)))
  (define-button 2 7 "Play Game" #f
   (lambda ()
    (message "")
    (play-game!)))
  (define-cycle-button 3 7 *game*
   (lambda () (message ""))
   (tic-tac-toe "Tic Tac Toe")
   (hexapawn "Hexapawn")
   (prolog "Prolog"))
  (define-button 4 7 "Play X" #f
   (lambda ()
    (message "")
    (play! 1)))
  (define-button 5 7 "Play O" #f
   (lambda ()
    (message "")
    (play! -1)))
  (define-button 6 7 "Clean Up" #f
   (lambda ()
    (message "")
    (achieve! (initial-board))))
  (define-button 7 7 "Set Prolog" #f
   (lambda ()
    (message "")
    (let* ((rules-pathname (format #f "~a/imitate/tool1/~a.pl"
				   (getenv "HOME") *pathname*))
	   (rules-list (read-file rules-pathname)))
      (set! *rules*
	    (append
	     (reduce append
		     (extract-clauses "initial_board" rules-list)
		     '())
	     (reduce append
		     (extract-clauses "legal_move" rules-list)
		     '())
	     (reduce append
		     (extract-clauses "outcome" rules-list)
		     '())))
      (prolog-command (format #f "consult('~a'),writeln('done')." rules-pathname))
      (redraw-display-pane))))
  ;; Ninth row
  (define-button 0 8 "Pick Up MP" #f
   (lambda ()
    (message "")
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (unless (and (= (number-of-fields *pathname*) 3)
		 (string->number (field-ref *pathname* 0))
		 (string->number (field-ref *pathname* 1)))
     (message "You need to supply coordinates")
     (abort))
    (pick-up-motor-program (vector (string->number (field-ref *pathname* 0))
				   (string->number (field-ref *pathname* 1))
				   (string->number (field-ref *pathname* 2)))
			   (first (string->list (field-ref *pathname* 3))))))
  (define-button 1 8 "Put Down MP" #f
   (lambda ()
    (message "")
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (unless (and (= (number-of-fields *pathname*) 3)
		 (string->number (field-ref *pathname* 0))
		 (string->number (field-ref *pathname* 1)))
     (message "You need to supply coordinates")
     (abort))
    (put-down-motor-program (vector (string->number (field-ref *pathname* 0))
				    (string->number (field-ref *pathname* 1))
				    (string->number (field-ref *pathname* 2)))
			    (first (string->list (field-ref *pathname* 3))))))
  (define-button 2 8 "Pick Up X" #f
   (lambda ()
    (message "")
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (unless *square* (message "No square selected") (abort))
    (pick-up-motor-program (square->position *square*) 'x)
    (set! *square* #f)))
  (define-button 3 8 "Pick Up O" #f
   (lambda ()
    (message "")
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (unless *square* (message "No square selected") (abort))
    (pick-up-motor-program (square->position *square*) 'o)
    (set! *square* #f)))
  (define-button 4 8 "Put Down X" #f
   (lambda ()
    (message "")
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (unless *square* (message "No square selected") (abort))
    (put-down-motor-program (square->position *square*) 'x)
    (set! *square* #f)))
  (define-button 5 8 "Put Down O" #f
   (lambda ()
    (message "")
    (unless *bluetooth-address*
     (message "No robot")
     (abort))
    (unless *square* (message "No square selected") (abort))
    (put-down-motor-program (square->position *square*) 'o)
    (set! *square* #f)))
  (define-button 6 8 "Find Board" #f
   (lambda ()
    (message "")
    (for-each-n (lambda (camera)
		 (hack-track-page-in-frames camera 0 1)
		 (hack-track-capture-frames camera 0 1 1))
		*number-of-cameras*)
    (find-board-in-pnm! (ppm->pgm (hack-track-frame->ppm *camera* 0)) *camera*)
    (set! *end* 0)
    (overlay-command)))
  (define-button 7 8 "Reconstruct" #f
   (lambda ()
    (message "")
    (for-each-n (lambda (camera)
		 (hack-track-page-in-frames camera 0 1)
		 (hack-track-capture-frames camera 0 1 1))
		*number-of-cameras*)
    (reconstruct-board-new! (hack-track-frame->ppm *camera* 0) *camera*)
    (set! *end* 0)
    (overlay-command)))
  ;; End of buttons
  (define-key (list (control #\x) (control #\c)) "Quit" quit)
  (define-key (control #\h) "Help"
   (lambda ()
    (for-each-n (lambda (camera) (hack-track-turn-off! camera))
		*number-of-cameras*)
    (help-command)))
  (define-key (control #\g) "Abort" abort-command)
  (set-background-task-enabler!
   (lambda ()
    (hack-track-unblock)
    (hack-track-set-interrupts?! #t)))
  (set-background-task-disabler!
   (lambda ()
    (hack-track-block)
    (hack-track-set-interrupts?! #f)))
  (hack-track-startup-with-viewfinder-multiple *number-of-cameras*
					       *window-xs*
					       *window-ys*
					       *height*
					       *width*
					       (buffer-size)
					       1)
  (recache-strings!)
  ;; debugging
  (when #f
   (for-each-n (lambda (camera) (v4l-set-agc! camera 0) #t)
	       *number-of-cameras*))
  (set! *squares*
	(list
	 ;; Left staging area
	 (create-square 20 320 'x '(staging left 0))
	 (create-square 20 360 'x '(staging left 1))
	 (create-square 20 400 'x '(staging left 2))
	 (create-square 20 440 'x '(staging left 3))
	 (create-square 20 480 'x '(staging left 4))
	 ;; 3*3 board
	 (create-square 100 360 #f '(board 0 0))
	 (create-square 100 400 #f '(board 1 0))
	 (create-square 100 440 #f '(board 2 0))
	 (create-square 140 360 #f '(board 0 1))
	 (create-square 140 400 #f '(board 1 1))
	 (create-square 140 440 #f '(board 2 1))
	 (create-square 180 360 #f '(board 0 2))
	 (create-square 180 400 #f '(board 1 2))
	 (create-square 180 440 #f '(board 2 2))
	 ;; Right staging area
	 (create-square 260 320 'o '(staging right 0))
	 (create-square 260 360 'o '(staging right 1))
	 (create-square 260 400 'o '(staging right 2))
	 (create-square 260 440 'o '(staging right 3))
	 (create-square 260 480 'o '(staging right 4)))))
 (lambda ()
  (for-each-n (lambda (camera) (hack-track-turn-on-viewfinder! camera 1))
	      *number-of-cameras*)
  (when *bluetooth-address*
   (park-arm)
   (wait-until-servos-stop)
   (relax-shoulder-elbow-and-head)))
 (lambda ()
  (when *bluetooth-address*
   (park-arm)
   (wait-until-servos-stop)
   (relax-shoulder-elbow-and-head))
  (hack-track-shutdown-multiple *number-of-cameras*))
 (lambda ()
  (set! *lines*
	(list (format #f "Base: ~s" *base*)
	      (format #f "Shoulder: ~s" *shoulder*)
	      (format #f "Elbow: ~s" *elbow*)
	      (format #f "Hand: ~s" *hand*)
	      (format #f "Left Finger: ~s" *left-finger*)
	      (format #f "Right Finger: ~s" *right-finger*)
	      (format #f "Head: ~s" *head*)))
  (for-each-n (lambda (camera)
	       (define-region
		(vector-ref *window-xs* camera)
		(vector-ref *window-ys* camera)
		*width*
		*height*
		(lambda (x1 y1)
		 (let* ((x1 (- x1 (vector-ref *window-xs* camera)))
			(y1 (- y1 (vector-ref *window-ys* camera)))
			(x1 (min x1 (- *width* 1)))
			(y1 (min y1 (- *height* 1))))
		  (message "")
		  (set! *camera* camera)
		  (message (format #f "(~s ~s)" x1 y1))
		  (redraw-display-pane)))))
	      *number-of-cameras*)
  (xclearwindow *display* *display-pane*)
  (redraw))
 (lambda ()
  (set! *pathname* (and (not (string=? *input* "")) *input*))
  (redraw-display-pane)))

(define-command (main
		 (at-most-one ("frame-rate"
			       frame-rate?
			       (frame-rate "n" real-argument 5.0)))
		 (at-most-one ("window-position"
			       window-position?
			       (window-position-x "x" integer-argument 0)
			       (window-position-y "y" integer-argument 0)))
		 (at-most-one ("sahelanthropus" sahelanthropus?)
			      ("tchadensis" tchadensis?)
			      ("orrorin" orrorin?)))
 (set! *frame-rate* (exact->inexact frame-rate))
 (hack-track-set-raw-frame-rate! frame-rate)
 (set! *number-of-cameras* 2)
 (set! *window-position?* #f)
 (when window-position?
  (set! *window-position?* #t)
  (set! *window-position-x* window-position-x)
  (set! *window-position-y* window-position-y))
 (set! *height* (+ (- 768 280) 105))
 (set! *width* (- 1024 10))
 (set! *who-am-i*
       (cond (sahelanthropus? 'sahelanthropus)
	     (tchadensis? 'tchadensis)
	     (orrorin? 'orrorin)
	     (else 'no-robot)))
 (set! *game* 'tic-tac-toe)
 (set! *game-variant* 'regular)
 (initialize-bluetooth-address!)
 (load-ellipse-finder-parameters!)
 (when *bluetooth-address* (load-robot-parameters!))
 (unless *bluetooth-address* (set! *reconstruct?* #f))
 ;; debugging
 (cond (#f
	(with-matlab-default-engine
	 (lambda (engine)
	  (set! *matlab-engine* engine)
	  (start-prolog!)
	  (when *bluetooth-address* (open-rfcomm! *bluetooth-address*))
	  (viewer '())
	  (when *bluetooth-address* (close-rfcomm!))
	  (stop-prolog!))))
       (else (set! *matlab-engine* #f)
	     (start-prolog!)
	     (when *bluetooth-address* (open-rfcomm! *bluetooth-address*))
	     (viewer '())
	     (when *bluetooth-address* (close-rfcomm!))
	     (stop-prolog!))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
