(MODULE TOOL1LIB-HACK-TRACK)
;;; LaHaShem HaAretz U'Mloah

;;; TTMTTD
;;;  1. way to startup just frame buffer, just frame buffer and viewfinder
;;;  2. (HACK-TRACK-SET-FRAME-FROM-PNM! i pnm)
;;;  3. (HACK-TRACK-SET-FRAMES-FROM-PNM-MOVIE! i j pnm-movie)
;;;  4. (HACK-TRACK-SET-FRAMES-FROM-VIDEO-FILE! i j pathname video-type)
;;;  5. make interchangeable between X and XShm
;;;  6. 24->32 bit so can use memcpy
;;;  7. double buffer, check for overrun
;;;  8. uncompressed single-file MBM, MGM, MPM file format
;;;       VIDEO-TYPE-ENUM
;;;       GET-VIDEO-WIDTH
;;;       GET-VIDEO-HEIGHT
;;;       READ-PPM-FROM-VIDEO-INPUT-PORT
;;;       READ-PIXMAP-FROM-VIDEO-INPUT-PORT
;;;       CLOSE-VIDEO-INPUT-PORT
;;;       OPEN-VIDEO-OUTPUT-FILE
;;;       WRITE-PNM-TO-VIDEO-OUTPUT-PORT
;;;       CLOSE-VIDEO-OUTPUT-PORT

(include "QobiScheme.sch")
(include "tool1lib-hack-track.sch")
(include "c-externals.sc")

;;; Structures

;;; Constants

(define wall-fixed #x100000000)
(define wall-fix 0.0)
(define wall-bad #t)

(define number-of-cameras 4)

;;; Variables

(define *hack-track-height* #f)
(define *hack-track-width* #f)
(define *hack-track-start* (vector #f #f #f #f))
(define *hack-track-finish* (vector #f #f #f #f))
(define *hack-track-raw-frame-rate* #f)
(define *red?* (vector #t #t #t #t))
(define *rhl* (vector 336 336 336 336))
(define *rhh* (vector 13 13 13 13))
(define *rsl* (vector 36 36 36 36))
(define *rsh* (vector 74 74 74 74))
(define *rvl* (vector 35 35 35 35))
(define *rvh* (vector 70 70 70 70))
(define *green?* (vector #t #t #t #t))
(define *ghl* (vector 122 122 122 122))
(define *ghh* (vector 164 164 164 164))
(define *gsl* (vector 29 29 29 29))
(define *gsh* (vector 71 71 71 71))
(define *gvl* (vector 22 22 22 22))
(define *gvh* (vector 55 55 55 55))
(define *blue?* (vector #t #t #t #t))
(define *bhl* (vector 192 192 192 192))
(define *bhh* (vector 227 227 227 227))
(define *bsl* (vector 32 32 32 32))
(define *bsh* (vector 66 66 66 66))
(define *bvl* (vector 27 27 27 27))
(define *bvh* (vector 61 61 61 61))
(define *yellow?* (vector #t #t #t #t))
(define *yhl* (vector 50 50 50 50))
(define *yhh* (vector 70 70 70 70))
(define *ysl* (vector 40 40 40 40))
(define *ysh* (vector 65 65 65 65))
(define *yvl* (vector 55 55 55 55))
(define *yvh* (vector 80 80 80 80))
(define *hack-track-motion-threshold* (vector 24 24 24 24))
(define *hack-track-colour-pixel-mass-threshold* (vector 60 60 60 60))
(define *hack-track-colour-proximity-threshold* (vector 4 4 4 4))
(define *hack-track-motion-pixel-mass-threshold* (vector 200 200 200 200))
(define *hack-track-motion-proximity-threshold* (vector 8 8 8 8))

;;; Procedures

(define (read-wall-clock)
 (let ((time (wall-clock)))
  (if (< time 0)
      (when wall-bad
       (set! wall-fix (+ wall-fixed wall-fix))
       (set! wall-bad #f))
      (set! wall-bad #t))
  (+ time wall-fix)))

(define (interruptible-usleep n)
 ;; needs work: This is hardwired to 100. It should be CLK_TCK. But we only
 ;;             expose CLOCKS_PER_SEC. So it goes.
 (hack-track-set-interrupts?! #t)
 (let ((end (+ (read-wall-clock) (* (/ n 1000000.0) 100))))
  (let loop ()
   (usleep 10000)
   (unless (> (read-wall-clock) end) (loop))))
 (hack-track-set-interrupts?! #f))

(define (v4l-available? camera) (= (v4l-available-internal camera) 1))

(define (vector-append . &rest)
 (list->vector (apply append (map vector->list &rest))))

(define (matrix-append . &rest) (apply map-vector vector-append &rest))

(define (debugging-hero p q r)
 (let* ((a (distance p q))
	(b (distance q r))
	(c (distance r p))
	(s (/ (+ a b c) 2)))
  (sqrt (max 0.0 (* s (- s a) (- s b) (- s c))))))

(define (debugging-area-of-convex-polygon points)
 (if (or (null? points) (null? (rest points)) (null? (rest (rest points))))
     0
     (let ((root (first points)))
      (let loop ((points (rest points)))
       (+ (debugging-hero root (first points) (second points))
	  (if (null? (rest (rest points))) 0 (loop (rest points))))))))

;;; Line Segments

(define (midpoint l)
 (vector (* 0.5 (+ (x (p l)) (x (q l)))) (* 0.5 (+ (y (p l)) (y (q l))))))

(define (point-of-intersection-of-lines l1 l2)
 ;; This will fail if the lines do not intersect or if they are collinear.
 (m*v (invert-matrix
       (vector (vector (- (y (p l1)) (y (q l1))) (- (x (q l1)) (x (p l1))))
	       (vector (- (y (p l2)) (y (q l2))) (- (x (q l2)) (x (p l2))))))
      (vector (+ (* (y (p l1)) (- (x (q l1)) (x (p l1))))
		 (* (x (p l1)) (- (y (p l1)) (y (q l1)))))
	      (+ (* (y (p l2)) (- (x (q l2)) (x (p l2))))
		 (* (x (p l2)) (- (y (p l2)) (y (q l2))))))))

(define (hack-track-set-raw-frame-rate! raw-frame-rate)
 (set! *hack-track-raw-frame-rate*
       (cond (raw-frame-rate (exact->inexact raw-frame-rate))
	     ((some-n v4l-available? number-of-cameras) 15.0)
	     (else (panic "No camera available")))))

(define (hack-track-using-v4l? camera)
 (not (zero? (hack-track-get-using-v4l-internal camera))))

(define (hack-track-interrupts?) (= (hack-track-interrupts-internal) 1))

(define (hack-track-set-interrupts?! on?)
 (hack-track-set-interrupts-internal! (if on? 1 0))
 #f)

(define (get-motion? camera i)
 (not (zero? (hack-track-get-motion-internal camera i))))

(define (hack-track-startup camera height width nframes subsample)
 (set! *hack-track-height* height)
 (set! *hack-track-width* width)
 (hack-track-startup-internal
  camera
  height
  width
  nframes
  subsample
  (inexact->exact (round *hack-track-raw-frame-rate*)))
 #f)

(define (hack-track-startup-offline camera height width nframes subsample)
 (set! *hack-track-height* height)
 (set! *hack-track-width* width)
 (hack-track-startup-offline-internal camera height width nframes subsample)
 #f)

(define (hack-track-startup-with-viewfinder
	 camera x y height width nframes subsample)
 (unless (and
	  (= (visual-class (xdefaultvisual *display* *screen*)) truecolor)
	  (= (visual-red_mask (xdefaultvisual *display* *screen*)) 16711680)
	  (= (visual-green_mask (xdefaultvisual *display* *screen*)) 65280)
	  (= (visual-blue_mask (xdefaultvisual *display* *screen*)) 255)
	  (= (xdefaultdepth *display* *screen*) 24))
  (panic "Hack-Track can only handle 24-bit TRUECOLOR visuals"))
 (set! *hack-track-height* height)
 (set! *hack-track-width* width)
 (hack-track-startup-with-viewfinder-internal
  camera
  height
  width
  nframes
  subsample
  (cdr *display*)
  *screen*
  *display-pane*
  x
  y
  *display-pane-width*
  *display-pane-height*
  (inexact->exact (round *hack-track-raw-frame-rate*)))
 #f)

(define (hack-track-restartup camera height width nframes subsample)
 (set! *hack-track-height* height)
 (set! *hack-track-width* width)
 (hack-track-restartup-internal
  camera
  height
  width
  nframes
  subsample
  (inexact->exact (round *hack-track-raw-frame-rate*)))
 #f)

(define (hack-track-turn-on-viewfinder! camera m)
 (hack-track-turn-on-viewfinder-internal! camera m)
 #f)

(define (hack-track-turn-on-capturing! camera i l n)
 (hack-track-turn-on-capturing-internal! camera i l n)
 #f)

(define (hack-track-turn-on-ellipse-tracker! camera i l n)
 (hack-track-turn-on-ellipse-tracker-internal! camera i l n)
 #f)

(define (hack-track-turn-on-ellipse-tracker-with-capturing! camera i l n)
 (hack-track-turn-on-ellipse-tracker-with-capturing-internal! camera i l n)
 #f)

(define (hack-track-turn-on-polygon-tracker! camera i l n h?)
 (hack-track-turn-on-polygon-tracker-internal! camera i l n (if h? 1 0))
 #f)

(define (hack-track-turn-on-polygon-tracker-with-capturing! camera i l n h?)
 (hack-track-turn-on-polygon-tracker-with-capturing-internal!
  camera i l n (if h? 1 0))
 #f)

(define (hack-track-turn-off! camera)
 (hack-track-turn-off-internal! camera)
 #f)

(define (hack-track-frame->ppm camera i)
 (let ((red (make-matrix *hack-track-height* *hack-track-width*))
       (green (make-matrix *hack-track-height* *hack-track-width*))
       (blue (make-matrix *hack-track-height* *hack-track-width*)))
  (do ((y 0 (+ y 1))) ((>= y *hack-track-height*))
   (do ((x 0 (+ x 1))) ((>= x *hack-track-width*))
    (matrix-set! red y x (hack-track-get-red camera i y x))
    (matrix-set! green y x (hack-track-get-green camera i y x))
    (matrix-set! blue y x (hack-track-get-blue camera i y x))))
  (make-ppm #t 255 red green blue)))

(define (hack-track-code-frame->ppm camera i)
 (let ((red (make-matrix *hack-track-height* *hack-track-width*))
       (green (make-matrix *hack-track-height* *hack-track-width*))
       (blue (make-matrix *hack-track-height* *hack-track-width*)))
  (do ((y 0 (+ y 1))) ((>= y *hack-track-height*))
   (do ((x 0 (+ x 1))) ((>= x *hack-track-width*))
    (matrix-set! red y x (hack-track-get-code-frame-red camera i y x))
    (matrix-set! green y x (hack-track-get-code-frame-green camera i y x))
    (matrix-set! blue y x (hack-track-get-code-frame-blue camera i y x))))
  (make-ppm #t 255 red green blue)))

(define (hack-track-frames->ppm-movie camera m n)
 (map-n-vector (lambda (i) (hack-track-frame->ppm camera (+ m i))) (- n m)))

(define (hack-track-code-frames->ppm-movie camera m n)
 (map-n-vector (lambda (i) (hack-track-code-frame->ppm camera (+ m i)))
	       (- n m)))

(define (ximage->ppm ximage height width)
 ;; This is hardwired to raw and 8x8x8 RGB.
 (make-ppm
  #t
  255
  (map-n-vector
   (lambda (y)
    (map-n-vector (lambda (x) (quotient (xgetpixel ximage x y) 65536)) width))
   height)
  (map-n-vector
   (lambda (y)
    (map-n-vector
     (lambda (x) (quotient (remainder (xgetpixel ximage x y) 65536) 256))
     width))
   height)
  (map-n-vector
   (lambda (y)
    (map-n-vector (lambda (x) (remainder (xgetpixel ximage x y) 256)) width))
   height)))

(define (hack-track! camera m n o)
 (hack-track-start! camera m)
 (for-each-n (lambda (i) (hack-track-step! camera (+ m i) (+ o i)) #f) (- n m))
 (hack-track-finish! camera)
 #f)

(define (hack-track-objects->pnm-movie camera m n o)
 (hack-track! camera m n o)
 (hack-track-frames->ppm-movie camera o (+ o (- n m))))

(define (hack-track-frame->pixmap camera i)
 (let* ((default-visual (xdefaultvisual *display* *screen*))
	(default-depth (xdefaultdepth *display* *screen*))
	(data (c-malloc (* 4 *hack-track-width* *hack-track-height*)))
	(ximage (cons 'ximagep
		      (c-xcreateimage
		       (cdr *display*) (cdr default-visual) default-depth
		       zpixmap 0 data *hack-track-width* *hack-track-height*
		       32 0)))
	(pixmap (xcreatepixmap
		 *display* *display-pane*
		 *hack-track-width* *hack-track-height* default-depth)))
  (hack-track-copy-frame-to-ximage! camera i (cdr ximage))
  (xputimage *display* pixmap *color-gc* ximage 0 0 0 0
	     *hack-track-width* *hack-track-height*)
  (xdestroyimage ximage)
  pixmap))

(define (hack-track-frames->pixmaps camera m n)
 (let* ((default-visual (xdefaultvisual *display* *screen*))
	(default-depth (xdefaultdepth *display* *screen*))
	(data (c-malloc (* 4 *hack-track-width* *hack-track-height*)))
	(ximage (cons 'ximagep
		      (c-xcreateimage
		       (cdr *display*) (cdr default-visual) default-depth
		       zpixmap 0 data *hack-track-width* *hack-track-height*
		       32 0)))
	(pixmaps (map-n-vector
		  (lambda (i)
		   (xcreatepixmap
		    *display* *display-pane*
		    *hack-track-width* *hack-track-height* default-depth))
		  (- n m))))
  (do ((i 0 (+ i 1))) ((>= i (- n m)))
   (hack-track-copy-frame-to-ximage! camera (+ m i) (cdr ximage))
   (xputimage *display* (vector-ref pixmaps i) *color-gc* ximage 0 0 0 0
	      *hack-track-width* *hack-track-height*))
  (xdestroyimage ximage)
  pixmaps))

(define (hack-track-motion->pixmap camera i)
 (let* ((default-visual (xdefaultvisual *display* *screen*))
	(default-depth (xdefaultdepth *display* *screen*))
	(data (c-malloc (* 4 *hack-track-width* *hack-track-height*)))
	(ximage (cons 'ximagep
		      (c-xcreateimage
		       (cdr *display*) (cdr default-visual) default-depth
		       zpixmap 0 data *hack-track-width* *hack-track-height*
		       32 0)))
	(pixmap (xcreatepixmap
		 *display* *display-pane*
		 *hack-track-width* *hack-track-height* default-depth)))
  (hack-track-motion-detector! camera i (cdr ximage))
  (xputimage *display* pixmap *color-gc* ximage 0 0 0 0
	     *hack-track-width* *hack-track-height*)
  (xdestroyimage ximage)
  pixmap))

(define (hack-track-motion->pixmaps camera m n)
 (let* ((default-visual (xdefaultvisual *display* *screen*))
	(default-depth (xdefaultdepth *display* *screen*))
	(data (c-malloc (* 4 *hack-track-width* *hack-track-height*)))
	(ximage (cons 'ximagep
		      (c-xcreateimage
		       (cdr *display*) (cdr default-visual) default-depth
		       zpixmap 0 data *hack-track-width* *hack-track-height*
		       32 0)))
	(pixmaps (map-n-vector
		  (lambda (i)
		   (xcreatepixmap
		    *display* *display-pane*
		    *hack-track-width* *hack-track-height* default-depth))
		  (- n m))))
  (do ((i 0 (+ i 1))) ((>= i (- n m)))
   ;; note: This is a kludge to deal with the fact that pairwise sequences
   ;;       are one shorter than the original.
   (hack-track-motion-detector! camera (+ m (min i (- n 2))) (cdr ximage))
   (xputimage *display* (vector-ref pixmaps i) *color-gc* ximage 0 0 0 0
	      *hack-track-width* *hack-track-height*))
  (xdestroyimage ximage)
  pixmaps))

(define (hack-track-objects->pixmaps camera m n o)
 (hack-track! camera m n o)
 (hack-track-frames->pixmaps camera o (+ o (- n m))))

(define (hack-track-show-frame camera i)
 (hack-track-show-frame-internal camera i)
 #f)

(define (hack-track-show-code-frame camera i)
 (hack-track-show-code-frame-internal camera i)
 #f)

(define (hack-track-show-frames camera m n)
 (do ((i m (+ i 1))) ((>= i n))
  (hack-track-show-frame camera i)
  (usleep (inexact->exact (round (/ 1000000.0 *frame-rate*))))))

(define (hack-track-set-thresholds! camera
				    motion-threshold
				    colour-pixel-mass-threshold
				    colour-proximity-threshold
				    motion-pixel-mass-threshold
				    motion-proximity-threshold)
 (hack-track-set-thresholds-internal! camera
				      motion-threshold
				      colour-pixel-mass-threshold
				      colour-proximity-threshold
				      motion-pixel-mass-threshold
				      motion-proximity-threshold)
 #f)

;;; begin need change

(define (hack-track-set-red-range! camera p? hl hh sl sh vl vh)
 (hack-track-set-red-range-internal! camera hl hh (if p? sl 101) sh vl vh)
 #f)

(define (hack-track-set-green-range! camera p? hl hh sl sh vl vh)
 (hack-track-set-green-range-internal! camera hl hh (if p? sl 101) sh vl vh)
 #f)

(define (hack-track-set-blue-range! camera p? hl hh sl sh vl vh)
 (hack-track-set-blue-range-internal! camera hl hh (if p? sl 101) sh vl vh)
 #f)

(define (hack-track-set-yellow-range! camera p? hl hh sl sh vl vh)
 (hack-track-set-yellow-range-internal! camera hl hh (if p? sl 101) sh vl vh)
 #f)

;;; end need change

(define (hack-track-capture-frames camera i l n)
 (hack-track-set-interrupts?! #t)
 (hack-track-capture-frames-internal camera i l n)
 (hack-track-set-interrupts?! #f))

(define (hack-track-frames->video-file camera m n pathname video-type)
 (call-with-video-output-file
  pathname video-type (- n m)
  (lambda (video-output-port)
   (do ((i m (+ i 1))) ((>= i n))
    (write-pnm-to-video-output-port
     (hack-track-frame->ppm camera i) video-output-port)))))

(define (hack-track->pnm-movie
	 camera before after height width nframes subsample)
 (let ((downsample-factor (/ *hack-track-raw-frame-rate* *frame-rate*)))
  (unless (< (abs (- downsample-factor (round downsample-factor))) 0.001)
   (panic "Frame rate does not yield an integral downsample factor"))
  (hack-track-startup camera height width nframes subsample)
  (before)
  (hack-track-capture-frames
   camera 0 (inexact->exact (round downsample-factor)) nframes)
  (after)
  (let ((pnm-movie (hack-track-frames->ppm-movie camera 0 nframes)))
   (hack-track-shutdown camera)
   pnm-movie)))

(define (hack-track->video-file camera
				before
				after
				height
				width
				nframes
				subsample
				pathname
				video-type)
 (let ((downsample-factor (/ *hack-track-raw-frame-rate* *frame-rate*)))
  (unless (< (abs (- downsample-factor (round downsample-factor))) 0.001)
   (panic "Frame rate does not yield an integral downsample factor"))
  (hack-track-startup camera height width nframes subsample)
  (before)
  (hack-track-capture-frames
   camera 0 (inexact->exact (round downsample-factor)) nframes)
  (after)
  (hack-track-frames->video-file camera 0 nframes pathname video-type)
  (hack-track-shutdown camera)
  #f))

(define (hack-track-ppm->frame! camera i ppm)
 (unless (and (= *hack-track-height* (pnm-height ppm))
	      (= *hack-track-width* (pnm-width ppm)))
  (panic "Wrong height or width"))
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (do ((y 0 (+ y 1))) ((>= y *hack-track-height*))
   (do ((x 0 (+ x 1))) ((>= x *hack-track-width*))
    (hack-track-set-red! camera i y x (matrix-ref red y x))
    (hack-track-set-green! camera i y x (matrix-ref green y x))
    (hack-track-set-blue! camera i y x (matrix-ref blue y x))))))

(define (get-ellipse-movie camera start end)
 (map-n-vector
  (lambda (frame)
   (map-n (lambda (k)
	   (make-ellipse (get-ellipse-x camera (+ start frame) k)
			 (get-ellipse-y camera (+ start frame) k)
			 (get-ellipse-theta camera (+ start frame) k)
			 (get-ellipse-r camera (+ start frame) k)
			 (get-ellipse-s camera (+ start frame) k)))
	  (get-object-k camera (+ start frame))))
  (- end start)))

(define (get-polygon-movie camera start finish)
 (map-n-vector
  (lambda (frame)
   (map-n
    (lambda (k)
     (map-n-vector
      (lambda (l)
       (vector (hack-track-get-polygon-x camera (+ start frame) k l)
	       (hack-track-get-polygon-y camera (+ start frame) k l)))
      (hack-track-get-polygon-l camera (+ start frame) k)))
    (get-object-k camera (+ start frame))))
  (- finish start)))

;;; begin need change

(define (hack-track-read-rgby-file! camera rgby? rgby)
 (when rgby?
  (let ((rgby (read-object-from-file (default-extension rgby "rgby"))))
   (vector-set! *red?* camera (first (first rgby)))
   (vector-set! *rhl* camera (second (first rgby)))
   (vector-set! *rhh* camera (third (first rgby)))
   (vector-set! *rsl* camera (fourth (first rgby)))
   (vector-set! *rsh* camera (fifth (first rgby)))
   (vector-set! *rvl* camera (sixth (first rgby)))
   (vector-set! *rvh* camera (seventh (first rgby)))
   (vector-set! *green?* camera (first (second rgby)))
   (vector-set! *ghl* camera (second (second rgby)))
   (vector-set! *ghh* camera (third (second rgby)))
   (vector-set! *gsl* camera (fourth (second rgby)))
   (vector-set! *gsh* camera (fifth (second rgby)))
   (vector-set! *gvl* camera (sixth (second rgby)))
   (vector-set! *gvh* camera (seventh (second rgby)))
   (vector-set! *blue?* camera (first (third rgby)))
   (vector-set! *bhl* camera (second (third rgby)))
   (vector-set! *bhh* camera (third (third rgby)))
   (vector-set! *bsl* camera (fourth (third rgby)))
   (vector-set! *bsh* camera (fifth (third rgby)))
   (vector-set! *bvl* camera (sixth (third rgby)))
   (vector-set! *bvh* camera (seventh (third rgby)))
   (vector-set! *yellow?* camera (first (fourth rgby)))
   (vector-set! *yhl* camera (second (fourth rgby)))
   (vector-set! *yhh* camera (third (fourth rgby)))
   (vector-set! *ysl* camera (fourth (fourth rgby)))
   (vector-set! *ysh* camera (fifth (fourth rgby)))
   (vector-set! *yvl* camera (sixth (fourth rgby)))
   (vector-set! *yvh* camera (seventh (fourth rgby)))
   (vector-set! *hack-track-motion-threshold* camera (first (fifth rgby)))
   (vector-set!
    *hack-track-colour-pixel-mass-threshold* camera (second (fifth rgby)))
   (vector-set!
    *hack-track-colour-proximity-threshold* camera (third (fifth rgby)))
   (vector-set!
    *hack-track-motion-pixel-mass-threshold* camera (fourth (fifth rgby)))
   (vector-set!
    *hack-track-motion-proximity-threshold* camera (fifth (fifth rgby)))))
 (hack-track-set-red-range! camera
			    (vector-ref *red?* camera)
			    (vector-ref *rhl* camera)
			    (vector-ref *rhh* camera)
			    (vector-ref *rsl* camera)
			    (vector-ref *rsh* camera)
			    (vector-ref *rvl* camera)
			    (vector-ref *rvh* camera))
 (hack-track-set-green-range! camera
			      (vector-ref *green?* camera)
			      (vector-ref *ghl* camera)
			      (vector-ref *ghh* camera)
			      (vector-ref *gsl* camera)
			      (vector-ref *gsh* camera)
			      (vector-ref *gvl* camera)
			      (vector-ref *gvh* camera))
 (hack-track-set-blue-range! camera
			     (vector-ref *blue?* camera)
			     (vector-ref *bhl* camera)
			     (vector-ref *bhh* camera)
			     (vector-ref *bsl* camera)
			     (vector-ref *bsh* camera)
			     (vector-ref *bvl* camera)
			     (vector-ref *bvh* camera))
 (hack-track-set-yellow-range! camera
			       (vector-ref *yellow?* camera)
			       (vector-ref *yhl* camera)
			       (vector-ref *yhh* camera)
			       (vector-ref *ysl* camera)
			       (vector-ref *ysh* camera)
			       (vector-ref *yvl* camera)
			       (vector-ref *yvh* camera))
 (hack-track-set-thresholds!
  camera
  (vector-ref *hack-track-motion-threshold* camera)
  (vector-ref *hack-track-colour-pixel-mass-threshold* camera)
  (vector-ref *hack-track-colour-proximity-threshold* camera)
  (vector-ref *hack-track-motion-pixel-mass-threshold* camera)
  (vector-ref *hack-track-motion-proximity-threshold* camera)))

(define (hack-track-write-rgby-file! camera pathname)
 (write-object-to-file
  (list (list (vector-ref *red?* camera)
	      (vector-ref *rhl* camera)
	      (vector-ref *rhh* camera)
	      (vector-ref *rsl* camera)
	      (vector-ref *rsh* camera)
	      (vector-ref *rvl* camera)
	      (vector-ref *rvh* camera))
	(list (vector-ref *green?* camera)
	      (vector-ref *ghl* camera)
	      (vector-ref *ghh* camera)
	      (vector-ref *gsl* camera)
	      (vector-ref *gsh* camera)
	      (vector-ref *gvl* camera)
	      (vector-ref *gvh* camera))
	(list (vector-ref *blue?* camera)
	      (vector-ref *bhl* camera)
	      (vector-ref *bhh* camera)
	      (vector-ref *bsl* camera)
	      (vector-ref *bsh* camera)
	      (vector-ref *bvl* camera)
	      (vector-ref *bvh* camera))
	(list (vector-ref *yellow?* camera)
	      (vector-ref *yhl* camera)
	      (vector-ref *yhh* camera)
	      (vector-ref *ysl* camera)
	      (vector-ref *ysh* camera)
	      (vector-ref *yvl* camera)
	      (vector-ref *yvh* camera))
	(list (vector-ref *hack-track-motion-threshold* camera)
	      (vector-ref *hack-track-colour-pixel-mass-threshold* camera)
	      (vector-ref *hack-track-colour-proximity-threshold* camera)
	      (vector-ref *hack-track-motion-pixel-mass-threshold* camera)
	      (vector-ref *hack-track-motion-proximity-threshold* camera)))
  (replace-extension pathname "rgby")))

;;; end need change

(define (hack-track-startup-multiple
	 number-of-cameras height width nframes subsample)
 (set! *hack-track-height* height)
 (set! *hack-track-width* width)
 (hack-track-startup-internal
  number-of-cameras
  height
  width
  nframes
  subsample
  (inexact->exact (round *hack-track-raw-frame-rate*)))
 #f)

(define (hack-track-startup-with-viewfinder-multiple
	 number-of-cameras xs ys height width nframes subsample)
 (unless (and
	  (= (visual-class (xdefaultvisual *display* *screen*)) truecolor)
	  (= (visual-red_mask (xdefaultvisual *display* *screen*)) 16711680)
	  (= (visual-green_mask (xdefaultvisual *display* *screen*)) 65280)
	  (= (visual-blue_mask (xdefaultvisual *display* *screen*)) 255)
	  (= (xdefaultdepth *display* *screen*) 24))
  (panic "Hack-Track can only handle 24-bit TRUECOLOR visuals"))
 (set! *hack-track-height* height)
 (set! *hack-track-width* width)
 (let ((xsc (make-c-int-vector number-of-cameras))
       (ysc (make-c-int-vector number-of-cameras)))
  (for-each-n (lambda (camera)
	       (c-int-vector-set! xsc camera (vector-ref xs camera))
	       (c-int-vector-set! ysc camera (vector-ref ys camera))
	       #f)
	      number-of-cameras)
  (hack-track-startup-with-viewfinder-multiple-internal
   number-of-cameras
   height
   width
   nframes
   subsample
   (cdr *display*)
   *screen*
   *display-pane*
   xsc
   ysc
   *display-pane-width*
   *display-pane-height*
   (inexact->exact (round *hack-track-raw-frame-rate*)))
  (free-c-int-vector! xsc)
  (free-c-int-vector! ysc)
  #f)
 #f)

(define (hack-track-restartup-multiple
	 number-of-cameras height width nframes subsample)
 (set! *hack-track-height* height)
 (set! *hack-track-width* width)
 (hack-track-restartup-internal
  number-of-cameras
  height
  width
  nframes
  subsample
  (inexact->exact (round *hack-track-raw-frame-rate*)))
 #f)

;;; Tam V'Nishlam Shevah L'El Borei Olam
